from __future__ import annotations

import json
from threading import Thread
from typing import TYPE_CHECKING

import logistro

from choreographer import protocol
from choreographer.channels import ChannelClosedError

if TYPE_CHECKING:
    from typing import Any

    from choreographer.browser_sync import BrowserSync
    from choreographer.channels._interface_type import ChannelInterface

_logger = logistro.getLogger(__name__)


class BrokerSync:
    """BrokerSync is a middleware implementation for synchronous browsers."""

    _browser: BrowserSync
    """Browser is a reference to the Browser object this broker is brokering for."""
    _channel: ChannelInterface
    """
    Channel will be the ChannelInterface implementation (pipe or websocket)
    that the broker communicates on.
    """

    def __init__(self, browser: BrowserSync, channel: ChannelInterface) -> None:
        """
        Construct a broker for a synchronous arragenment w/ both ends.

        Args:
            browser: The sync browser implementation.
            channel: The channel the browser uses to talk on.

        """
        self._browser = browser
        self._channel = channel

    def run_output_thread(self, **kwargs: Any) -> None:
        """
        Run a thread which dumps all browser messages. kwargs is passed to print.

        Raises:
            ChannelClosedError: When the channel is closed, this error is raised.

        """

        def run_print() -> None:
            try:
                while True:
                    responses = self._channel.read_jsons()
                    for response in responses:
                        print(json.dumps(response, indent=4), **kwargs)  # noqa: T201 print in the point
            except ChannelClosedError:
                print("ChannelClosedError caught.", **kwargs)  # noqa: T201 print is the point

        _logger.info("Starting thread to dump output to stdout.")
        Thread(target=run_print).start()

    def write_json(self, obj: protocol.BrowserCommand) -> protocol.MessageKey | None:
        """
        Send an object down the channel.

        Args:
            obj: An object to be serialized to json and written to the channel.

        """
        protocol.verify_params(obj)
        key = protocol.calculate_message_key(obj)
        self._channel.write_json(obj)
        return key

    def clean(self) -> None:
        pass



"""Provides the basic protocol class (the abstract base) for a protocol."""

from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any, Mapping, MutableMapping, Sequence

    from choreographer.channels._interface_type import ChannelInterface


class BrowserImplInterface(Protocol):
    """Defines the basic interface of a channel."""

    # I guess we need to include __init__?
    def __init__(
        self,
        channel: ChannelInterface,
        path: Path | str | None = None,
        **kwargs: Any,
    ) -> None: ...
    def pre_open(self) -> None: ...
    def get_popen_args(self) -> Mapping[str, Any]: ...
    def get_cli(self) -> Sequence[str]: ...
    def get_env(self) -> MutableMapping[str, str]: ...
    def clean(self) -> None: ...
    def is_isolated(self) -> bool: ...



from __future__ import annotations

import os
import platform
import shutil
import stat
import sys
import tempfile
import time
import warnings
from pathlib import Path
from typing import TYPE_CHECKING

import logistro

if TYPE_CHECKING:
    from typing import Any, Callable, MutableMapping, Sequence

_logger = logistro.getLogger(__name__)


class TmpDirWarning(UserWarning):
    """A warning if for whatever reason we can't eliminate the tmp dir."""


class TmpDirectory:
    """
    The python stdlib `TemporaryDirectory` wrapper for easier use.

    Python's `TemporaryDirectory` suffered a couple API changes that mean
    you can't call it the same way for similar versions. This wrapper is
    also much more aggressive about deleting the directory when it's done,
    not necessarily relying on OS functions.
    """

    temp_dir: tempfile.TemporaryDirectory[str]
    """A reference to the underlying python `TemporaryDirectory` implementation."""
    path: Path
    """The path to the temporary directory."""
    exists: bool
    """A flag to indicate if the directory still exists."""

    def __init__(self, path: str | None = None, *, sneak: bool = False):
        """
        Construct a wrapped `TemporaryDirectory`.

        Args:
            path: manually specify the directory to use
            sneak: (default False) avoid using /tmp
                Ubuntu's snap will sandbox /tmp

        """
        self._with_onexc = bool(sys.version_info[:3] >= (3, 12))
        args: MutableMapping[str, Any] = {}

        if path:
            args = {"dir": path}
        elif sneak:
            args = {"prefix": ".choreographer-", "dir": Path.home()}

        if platform.system() != "Windows":
            self.temp_dir = tempfile.TemporaryDirectory(**args)
        else:  # is windows
            vinfo = sys.version_info[:3]
            if vinfo >= (3, 12):
                self.temp_dir = tempfile.TemporaryDirectory(  # type: ignore [call-overload, unused-ignore]
                    delete=False,
                    ignore_cleanup_errors=True,
                    **args,
                )
            elif vinfo >= (3, 10):
                self.temp_dir = tempfile.TemporaryDirectory(  # type: ignore [call-overload, unused-ignore]
                    ignore_cleanup_errors=True,
                    **args,
                )
            else:
                self.temp_dir = tempfile.TemporaryDirectory(**args)

        self.path = Path(self.temp_dir.name)
        _logger.info(f"Temp directory created: {self.path}.")
        self.exists = True

    def _delete_manually(  # noqa: C901, PLR0912
        self,
        *,
        check_only: bool = False,
        quiet: bool = False,
    ) -> tuple[
        int,
        int,
        Sequence[tuple[Path, BaseException]],
    ]:
        if not self.path.exists():
            self.exists = False
            return 0, 0, []
        n_dirs = 0
        n_files = 0
        errors = []
        for root, dirs, files in os.walk(self.path, topdown=False):
            n_dirs += len(dirs)
            n_files += len(files)
            if not check_only:
                for f in files:
                    fp = Path(root) / f
                    _logger.debug2(f"Have file {fp}")
                    try:
                        fp.chmod(stat.S_IWUSR)
                        fp.unlink(missing_ok=True)
                        _logger.debug2("Deleted")
                    except BaseException as e:  # noqa: BLE001 yes catch and report
                        errors.append((fp, e))
                for d in dirs:
                    fp = Path(root) / d
                    _logger.debug2(f"Have directory {fp}")
                    try:
                        fp.chmod(stat.S_IWUSR)
                        fp.rmdir()
                        _logger.debug2("Deleted")
                    except BaseException as e:  # noqa: BLE001 yes catch and report
                        errors.append((fp, e))

            # clean up directory
        if not check_only:
            try:
                self.path.chmod(stat.S_IWUSR)
                self.path.rmdir()
            except BaseException as e:  # noqa: BLE001 yes catch and report
                errors.append((self.path, e))

        if check_only:
            if n_dirs or n_files:
                self.exists = True
            else:
                self.exists = False
        elif errors:
            if not quiet:
                warnings.warn(  # noqa: B028
                    "The temporary directory could not be deleted, "
                    f"execution will continue. errors: {errors}",
                    TmpDirWarning,
                )
            self.exists = True
        else:
            self.exists = False

        return n_dirs, n_files, errors

    def clean(self) -> None:  # noqa: C901
        """Try several different ways to eliminate the temporary directory."""
        try:
            # no faith in this python implementation, always fails with windows
            # very unstable recently as well, lots new arguments in tempfile package
            if hasattr(self, "temp_dir") and self.temp_dir:
                self.temp_dir.cleanup()
            self.exists = False
            _logger.info("TemporaryDirectory.cleanup() worked.")
        except BaseException as e:  # noqa: BLE001 we try many ways to clean, this is the first one
            _logger.info(f"TemporaryDirectory.cleanup() failed. Error {e}")

        # bad typing but tough
        def remove_readonly(
            func: Callable[[str], None],
            path: str | Path,
            _excinfo: Any,
        ) -> None:
            try:
                Path(path).chmod(stat.S_IWUSR)
                func(str(path))
            except FileNotFoundError:
                pass

        try:
            if self._with_onexc:
                shutil.rmtree(self.path, onexc=remove_readonly)  # type: ignore [call-arg, unused-ignore]
            else:
                shutil.rmtree(self.path, onerror=remove_readonly)
            self.exists = False
            if hasattr(self, "temp_dir"):
                del self.temp_dir
            _logger.info("shutil.rmtree worked.")
        except FileNotFoundError:
            self.exists = False
            if hasattr(self, "temp_dir"):
                del self.temp_dir
            _logger.info("shutil.rmtree worked.")
        except BaseException as e:  # noqa: BLE001
            _logger.debug("Error during tmp file removal.", exc_info=e)
            self._delete_manually(check_only=True)
            if not self.exists:
                return
            _logger.info(f"shutil.rmtree() failed to delete temporary file. Error {e}")

            def extra_clean() -> None:
                i = 0
                tries = 4
                while self.path.exists() and i < tries:
                    _logger.info(f"Extra manual clean executing {i}.")
                    self._delete_manually(quiet=True)
                    i += 1
                    time.sleep(2)
                if self.path.exists():
                    self._delete_manually(quiet=False)

            # testing doesn't look threads so I guess we'll block
            extra_clean()
            if self.path.exists():
                _logger.warning("Temporary dictory couldn't be removed manually.")



import warnings

import logistro
import pytest
import pytest_asyncio

import choreographer as choreo

# allows to create a browser pool for tests
pytestmark = pytest.mark.asyncio(loop_scope="function")

_logger = logistro.getLogger(__name__)


@pytest_asyncio.fixture(scope="function", loop_scope="function")
async def session(browser):
    _logger.info("testing...")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", choreo.protocol.ExperimentalFeatureWarning)
        session_browser = await browser.create_session()
    yield session_browser
    await browser.close_session(session_browser)


@pytest.mark.asyncio
async def test_session_send_command(session):
    _logger.info("testing...")
    # Test valid request with correct command
    response = await session.send_command(command="Target.getTargets")
    assert "result" in response and "targetInfos" in response["result"]  # noqa: PT018 I like this assertion

    # Test invalid method name should return error
    response = await session.send_command(command="dkadklqwmd")
    assert "error" in response

    # Test int method should return error
    with pytest.raises(
        choreo.protocol.MessageTypeError,
    ):
        await session.send_command(command=12345)



import asyncio
import logging

import logistro
import pytest
import pytest_asyncio

import choreographer as choreo
from choreographer import errors

_logger = logistro.getLogger(__name__)


@pytest.fixture(params=[True, False], ids=["enable_sandbox", ""])
def sandbox(request):
    return request.param


@pytest.fixture(params=[True, False], ids=["enable_gpu", ""])
def gpu(request):
    return request.param


@pytest.fixture(params=[True, False], ids=["headless", ""])
def headless(request):
    return request.param


# --headless is the default flag for most tests,
# but you can set --no-headless if you want to watch
def pytest_addoption(parser):
    parser.addoption("--headless", action="store_true", dest="headless", default=True)
    parser.addoption("--no-headless", dest="headless", action="store_false")


# browser fixture will supply a browser for you
@pytest_asyncio.fixture(scope="function", loop_scope="function")
async def browser(request):
    _logger.info("Fixture building browser.")
    headless = request.config.getoption("--headless")
    browser = await choreo.Browser(
        headless=headless,
    )
    yield browser
    _logger.info("Fixture closing browser.")
    try:
        await browser.close()
    except errors.BrowserClosedError:
        pass
    if browser._browser_impl.tmp_dir.exists:  # noqa: SLF001
        raise RuntimeError(
            "Temporary directory not deleted successfully: "
            f"{browser._browser_impl.tmp_dir.path}",  # noqa: SLF001
        )


# add a timeout if tests requests browser
# but if tests creates their own browser they are responsible
# a fixture can be used to specify the timeout: timeout=10
# else it uses pytest.default_timeout
@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_setup(item):
    # not even sure if this working
    # typer doesn't like item: pytest.Item w/ funcargs
    # but this is the recommended way
    # some people say do trylast
    yield
    if "browser" in item.funcargs:
        _logger.info("Hook setting test timeout.")
        raw_test_fn = item.obj
        timeouts = [k for k in item.funcargs if k.startswith("timeout")]
        timeout = (
            item.funcargs[timeouts[-1]] if len(timeouts) else pytest.default_timeout
        )
        if (
            item.get_closest_marker("asyncio") and timeout
        ):  # "closest" because markers can be function/session/package etc

            async def wrapped_test_fn(*args, **kwargs):
                try:
                    return await asyncio.wait_for(
                        raw_test_fn(*args, **kwargs),
                        timeout=timeout,
                    )
                except TimeoutError:
                    pytest.fail(
                        f"Test {item.name} failed a timeout. "
                        "This can be extended, but shouldn't be. See conftest.py.",
                    )

            item.obj = wrapped_test_fn


def pytest_configure():
    # change this by command line TODO
    pytest.default_timeout = 20


# pytest shuts down its capture before logging/threads finish
@pytest.fixture(scope="session", autouse=True)
def cleanup_logging_handlers(request):
    capture = request.config.getoption("--capture") != "no"
    try:
        yield
    finally:
        if capture:
            _logger.info("Conftest cleaning up handlers.")
            for handler in logging.root.handlers[:]:
                handler.flush()
                if isinstance(handler, logging.StreamHandler):
                    logging.root.removeHandler(handler)



"""A compilation of the errors available in choreographer."""

from ._brokers._async import UnhandledMessageWarning
from .browsers import (
    BrowserClosedError,
    BrowserDepsError,
    BrowserFailedError,
    ChromeNotFoundError,
)
from .channels import BlockWarning, ChannelClosedError
from .protocol import (
    DevtoolsProtocolError,
    ExperimentalFeatureWarning,
    MessageTypeError,
    MissingKeyError,
)
from .utils import TmpDirWarning

__all__ = [
    "BlockWarning",
    "BrowserClosedError",
    "BrowserDepsError",
    "BrowserFailedError",
    "ChannelClosedError",
    "ChromeNotFoundError",
    "DevtoolsProtocolError",
    "ExperimentalFeatureWarning",
    "MessageTypeError",
    "MissingKeyError",
    "TmpDirWarning",
    "UnhandledMessageWarning",
]



"""Provides a class proving tools for running chromium browsers."""

from __future__ import annotations

import os
import platform
import re
import subprocess
import sys
from pathlib import Path
from typing import TYPE_CHECKING

import logistro

if platform.system() == "Windows":
    import msvcrt

from choreographer.channels import Pipe
from choreographer.utils import TmpDirectory, get_browser_path

from ._chrome_constants import chrome_names, typical_chrome_paths

if TYPE_CHECKING:
    import logging
    from typing import Any, Mapping, MutableMapping, Sequence

    from choreographer.channels._interface_type import ChannelInterface

_chromium_wrapper_path = (
    Path(__file__).resolve().parent / "_unix_pipe_chromium_wrapper.py"
)

_packaged_chromium_libs = Path(__file__).resolve().parent / "packaged_chromium_libs"

_logger = logistro.getLogger(__name__)


def _is_exe(path: str | Path) -> bool:
    try:
        return os.access(path, os.X_OK)
    except:  # noqa: E722 bare except ok, weird errors, best effort.
        return False


_logs_parser_regex = re.compile(r"\d*:\d*:\d*\/\d*\.\d*:")


class ChromeNotFoundError(RuntimeError):
    """Raise when browser path can't be determined."""


class Chromium:
    """
    Chromium represents an implementation of the chromium browser.

    It also includes chromium-like browsers (chrome, edge, and brave).
    """

    path: str | Path | None
    """The path to the chromium executable."""
    gpu_enabled: bool
    """True if we should use the gpu. False by default for compatibility."""
    headless: bool
    """True if we should not show the browser, true by default."""
    sandbox_enabled: bool
    """True to enable the sandbox. False by default."""
    skip_local: bool
    """True if we want to avoid looking for our local download when searching path."""
    tmp_dir: TmpDirectory
    """A reference to a temporary directory object the chromium needs to store data."""

    @classmethod
    def logger_parser(
        cls,
        record: logging.LogRecord,
        _old: MutableMapping[str, Any],
    ) -> bool:
        """
        Parse (via `logging.Filter.parse()`) data from browser stderr for logging.

        Args:
            record: the `logging.LogRecord` object to read/modify
            _old: data that was already stripped out.

        """
        record.msg = _logs_parser_regex.sub("", record.msg)
        # we just eliminate their stamp, we dont' extract it
        return True

    def _libs_ok(self) -> bool:
        """Return true if libs ok."""
        if self.skip_local:
            _logger.debug(
                "If we HAVE to skip local.",
            )
            return True
        _logger.debug("Checking for libs needed.")
        if platform.system() != "Linux":
            _logger.debug("We're not in linux, so no need for check.")
            return True
        p = None
        try:
            _logger.debug(f"Trying ldd {self.path}")
            p = subprocess.run(  # noqa: S603, validating run with variables
                [  # noqa: S607 path is all we have
                    "ldd",
                    str(self.path),
                ],
                capture_output=True,
                timeout=5,
                check=True,
            )
        except Exception as e:  # noqa: BLE001
            msg = "ldd failed."
            stderr = p.stderr.decode() if p and p.stderr else None
            # Log failure as INFO rather than WARNING so that it's hidden by default,
            # since browser may succeed even if ldd fails
            _logger.info(
                msg  # noqa: G003 + in log
                + f" e: {e}, stderr: {stderr}",
            )
            return False
        if b"not found" in p.stdout:
            msg = "Found deps missing in chrome"
            _logger.debug2(msg + f" {p.stdout.decode()}")
            return False
        _logger.debug("No problems found with dependencies")
        return True

    def __init__(
        self,
        channel: ChannelInterface,
        path: Path | str | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Construct a chromium browser implementation.

        Args:
            channel: the `choreographer.Channel` we'll be using (WebSockets? Pipe?)
            path: path to the browser
            kwargs:
                gpu_enabled (default False): Turn on GPU? Doesn't work in all envs.
                headless (default True): Actually launch a browser?
                sandbox_enabled (default False): Enable sandbox-
                    a persnickety thing depending on environment, OS, user, etc
                tmp_dir (default None): Manually set the temporary directory

        Raises:
            RuntimeError: Too many kwargs, or browser not found.
            NotImplementedError: Pipe is the only channel type it'll accept right now.

        """
        _logger.info(f"Chromium init'ed with kwargs {kwargs}")
        self.path = path
        self.gpu_enabled = kwargs.pop("enable_gpu", False)
        self.headless = kwargs.pop("headless", True)
        self.sandbox_enabled = kwargs.pop("enable_sandbox", False)
        self._tmp_dir_path = kwargs.pop("tmp_dir", None)
        if kwargs:
            raise RuntimeError(
                f"Chromium.get_cli() received invalid args: {kwargs.keys()}",
            )
        self.skip_local = bool(
            "ubuntu" in platform.version().lower() and self.sandbox_enabled,
        )
        if self.skip_local:
            _logger.warning(
                "Skipping local. Ubuntu + Sandbox require using package manager.",
            )

        if not self.path:
            self.path = get_browser_path(
                executable_names=chrome_names,
                skip_local=self.skip_local,
            )
        if not self.path and typical_chrome_paths:
            # do typical chrome paths
            for candidate in typical_chrome_paths:
                if _is_exe(candidate):
                    self.path = candidate
                    break
        if not self.path:
            raise ChromeNotFoundError(
                "Browser not found. You can use get_chrome(), "
                "please see documentation.",
            )
        _logger.info(f"Found chromium path: {self.path}")

        self._channel = channel
        if not isinstance(channel, Pipe):
            raise NotImplementedError("Websocket style channels not implemented yet.")

        self._is_isolated = "snap" in str(self.path)

    def pre_open(self) -> None:
        """Prepare browser for opening."""
        self.tmp_dir = TmpDirectory(
            path=self._tmp_dir_path,
            sneak=self._is_isolated,
        )
        self.missing_libs = not self._libs_ok()
        _logger.info(f"Temporary directory at: {self.tmp_dir.path}")

    def is_isolated(self) -> bool:
        """
        Return if /tmp directory is isolated by OS.

        Returns:
            bool indicating if /tmp is isolated.

        """
        return self._is_isolated

    def get_popen_args(self) -> Mapping[str, Any]:
        """Return the args needed to runc chromium with `subprocess.Popen()`."""
        args = {}
        # need to check pipe
        if platform.system() == "Windows":
            args["creationflags"] = subprocess.CREATE_NEW_PROCESS_GROUP  # type: ignore [attr-defined]
            args["close_fds"] = False
        else:
            args["close_fds"] = True
            if isinstance(self._channel, Pipe):
                args["stdin"] = self._channel.from_choreo_to_external
                args["stdout"] = self._channel.from_external_to_choreo
        _logger.debug(f"Returning args: {args}")
        return args

    def get_cli(self) -> Sequence[str]:
        """Return the CLI command for chromium."""
        if platform.system() != "Windows":
            cli = [
                str(sys.executable),
                str(_chromium_wrapper_path),
                str(self.path),
            ]
        else:
            cli = [
                str(self.path),
            ]

        if not self.gpu_enabled:
            cli.append("--disable-gpu")
        if self.headless:
            cli.append("--headless")
        if not self.sandbox_enabled:
            cli.append("--no-sandbox")

        cli.extend(
            [
                "--disable-breakpad",
                "--allow-file-access-from-files",
                "--enable-logging=stderr",
                f"--user-data-dir={self.tmp_dir.path}",
                "--no-first-run",
                "--enable-unsafe-swiftshader",
                "--disable-dev-shm-usage",
                "--disable-background-media-suspend",
                "--disable-lazy-loading",
                "--disable-background-timer-throttling",
                "--disable-backgrounding-occluded-windows",
                "--disable-renderer-backgrounding",
                "--disable-component-update",
                "--disable-hang-monitor",
                "--disable-popup-blocking",
                "--disable-prompt-on-repost",
                "--disable-ipc-flooding-protection",
                "--disable-sync",
                "--metrics-recording-only",
                "--password-store=basic",
                "--use-mock-keychain",
                "--no-default-browser-check",
                "--no-process-per-site",
                "--disable-web-security",
            ],
        )
        if isinstance(self._channel, Pipe):
            cli.append("--remote-debugging-pipe")
            if platform.system() == "Windows":
                # its gonna read on 3
                # its gonna write on 4
                r_handle = msvcrt.get_osfhandle(self._channel.from_choreo_to_external)  # type: ignore [attr-defined]
                w_handle = msvcrt.get_osfhandle(self._channel.from_external_to_choreo)  # type: ignore [attr-defined]
                _inheritable = True
                os.set_handle_inheritable(r_handle, _inheritable)  # type: ignore [attr-defined]
                os.set_handle_inheritable(w_handle, _inheritable)  # type: ignore [attr-defined]
                cli += [
                    f"--remote-debugging-io-pipes={r_handle!s},{w_handle!s}",
                ]
        _logger.debug(f"Returning cli: {cli}")
        return cli

    def get_env(self) -> MutableMapping[str, str]:
        """Return the env needed for chromium."""
        env = os.environ.copy()
        return env

    def clean(self) -> None:
        """Clean up any leftovers form browser, like tmp files."""
        if hasattr(self, "tmp_dir"):
            self.tmp_dir.clean()

    def __del__(self) -> None:
        """Delete the temporary file and run `clean()`."""
        self.clean()



from __future__ import annotations

import platform
import subprocess

import logistro

_logger = logistro.getLogger(__name__)


def kill(process: subprocess.Popen[bytes]) -> None:
    if platform.system() == "Windows":
        subprocess.call(  # noqa: S603, false positive, input fine
            ["taskkill", "/F", "/T", "/PID", str(process.pid)],  # noqa: S607 windows full path...
            stderr=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
        )
    else:
        process.terminate()
        _logger.debug("Called terminate (a light kill).")
        if process.poll() is None:
            _logger.debug("Calling kill (a heavy kill).")
            process.kill()



"""Provides the basic protocol class (the abstract interface) for a channel."""

from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from typing import Any, Mapping, Sequence

    from choreographer.protocol import BrowserResponse


class ChannelInterface(Protocol):
    """Defines the basic interface of a channel."""

    # Not sure I like the obj type
    def write_json(self, obj: Mapping[str, Any]) -> None:
        """
        Accept an object and send it doesnt the channel serialized.

        Args:
            obj: the object to send to the browser.

        """

    def read_jsons(self, *, blocking: bool = True) -> Sequence[BrowserResponse]:
        """
        Read all available jsons in the channel and returns a list of complete ones.

        Args:
            blocking: should this method block on read or return immediately.

        """

    def close(self) -> None:
        """Close the channel."""



class BrowserClosedError(RuntimeError):
    """An error for when the browser is closed accidently (during access)."""


class BrowserFailedError(RuntimeError):
    """An error for when the browser fails to launch."""


class BrowserDepsError(BrowserFailedError):
    """An error for when the browser is closed because of missing libs."""

    def __init__(self):
        msg = (
            "It seems like you are running a slim version of your "
            "operating system and are missing some common dependencies. "
            "The following command should install the required "
            "dependencies on most systems:\n"
            "\n"
            "$ sudo apt update && sudo apt-get install libnss3 "
            "libatk-bridge2.0-0 libcups2 libxcomposite1 libxdamage1 "
            "libxfixes3 libxrandr2 libgbm1 libxkbcommon0 libpango-1.0-0 "
            "libcairo2 libasound2\n"
            "\n"
            "If you have already run the above command and are still "
            "seeing this error, or the above command fails, consult the "
            "Kaleido documentation for operating system to install "
            "chromium dependencies.\n"
            "\n"
            "For support, run the command `choreo_diagnose` and create "
            "an issue with its output."
        )
        super().__init__(msg)


# BrowserDeps being a more specific type of Failure.
# And Closed not necessarily being related (you can intentionally closed,
# and something else can error because its closed.)



"""Provide a lower-level async interface to the Devtools Protocol."""

from __future__ import annotations

import inspect
from typing import TYPE_CHECKING

import logistro

from choreographer import protocol

if TYPE_CHECKING:
    import asyncio
    from typing import Any, Callable, Coroutine, MutableMapping

    from choreographer._brokers import Broker

_logger = logistro.getLogger(__name__)


class Session:
    """A session is a single conversation with a single target."""

    session_id: str
    """The id of the session given by the browser."""
    message_id: int
    """All messages are counted per session and this is the current message id."""
    subscriptions: MutableMapping[
        str,
        tuple[
            Callable[[protocol.BrowserResponse], Coroutine[Any, Any, Any]],
            bool,
        ],
    ]

    def __init__(self, session_id: str, broker: Broker) -> None:
        """
        Construct a session from the browser as an object.

        A session is like an open conversation with a target.
        All commands are sent on sessions.

        Args:
            broker:  a reference to the browser's broker
            session_id:  the id given by the browser

        """
        if not isinstance(session_id, str):
            raise TypeError("session_id must be a string")
        # Resources
        self._broker = broker

        # State
        self.session_id = session_id
        _logger.debug(f"New session: {session_id}")
        self.message_id = 0
        self.subscriptions = {}

    async def send_command(
        self,
        command: str,
        params: MutableMapping[str, Any] | None = None,
    ) -> protocol.BrowserResponse:
        """
        Send a devtools command on the session.

        https://chromedevtools.github.io/devtools-protocol/

        Args:
            command: devtools command to send
            params: the parameters to send

        Returns:
            A message key (session, message id) tuple or None

        """
        current_id = self.message_id
        self.message_id += 1
        json_command = protocol.BrowserCommand(
            {
                "id": current_id,
                "method": command,
            },
        )

        if self.session_id:
            json_command["sessionId"] = self.session_id
        if params:
            json_command["params"] = params
        _logger.debug(
            f"Cmd '{command}', param keys '{params.keys() if params else ''}', "
            f"sessionId '{self.session_id}'",
        )
        _logger.debug2(f"Full params: {str(params).replace('%', '%%')}")
        return await self._broker.write_json(json_command)

    def subscribe(
        self,
        string: str,
        callback: Callable[[protocol.BrowserResponse], Coroutine[Any, Any, Any]],
        *,
        repeating: bool = True,
    ) -> None:
        """
        Subscribe to an event on this session.

        Args:
            string: the name of the event. Can use * wildcard at the end.
            callback: the callback (which takes a message dict and returns nothing)
            repeating: default True, should the callback execute more than once

        """
        if not inspect.iscoroutinefunction(callback):
            raise TypeError(
                "Call back must be be `async def` type function.",
            )
        if string in self.subscriptions:
            raise ValueError(
                "You are already subscribed to this string, "
                "duplicate subscriptions are not allowed.",
            )
        else:
            # so this should be per session
            # and that means we need a list of all sessions
            self.subscriptions[string] = (callback, repeating)

    def unsubscribe(self, string: str) -> None:
        """
        Remove a subscription.

        Args:
            string: the subscription to remove.

        """
        if string not in self.subscriptions:
            return
        del self.subscriptions[string]

    def subscribe_once(self, string: str) -> asyncio.Future[Any]:
        """
        Return a future for a browser event.

        Generally python asyncio doesn't recommend futures.

        But in this case, one must call subscribe_once and await it later,
        generally because they must subscribe and then provoke the event.

        Args:
            string: the event to subscribe to

        Returns:
            A future to be awaited later, the complete event.

        """
        return self._broker.new_subscription_future(self.session_id, string)


class Target:
    """A target like a browser, tab, or others. It sends commands. It has sessions."""

    target_id: str
    """The browser's ID of the target."""
    sessions: MutableMapping[str, Session]
    """A list of all the sessions for this target."""

    def __init__(self, target_id: str, broker: Broker):
        """
        Create a target after one ahs been created by the browser.

        Args:
            broker:  a reference to the browser's broker
            target_id:  the id given by the browser

        """
        if not isinstance(target_id, str):
            raise TypeError("target_id must be string")
        # Resources
        self._broker = broker

        # States
        self.sessions = {}
        self.target_id = target_id
        _logger.debug(f"Created new target {target_id}.")

    def _add_session(self, session: Session) -> None:
        if not isinstance(session, Session):
            raise TypeError("session must be a session type class")
        self.sessions[session.session_id] = session

    def _remove_session(self, session_id: str) -> None:
        if isinstance(session_id, Session):
            session_id = session_id.session_id
        _ = self.sessions.pop(session_id, None)

    def get_session(self) -> Session:
        """Retrieve the first session of the target, if it exists."""
        if not self.sessions.values():
            raise RuntimeError(
                "Cannot use this method without at least one valid session",
            )
        session = next(iter(self.sessions.values()))
        return session

    async def send_command(
        self,
        command: str,
        params: MutableMapping[str, Any] | None = None,
    ) -> protocol.BrowserResponse:
        """
        Send a command to the first session in a target.

        https://chromedevtools.github.io/devtools-protocol/

        Args:
            command: devtools command to send
            params: the parameters to send

        """
        if not self.sessions.values():
            raise RuntimeError("Cannot send_command without at least one valid session")
        session = self.get_session()
        return await session.send_command(command, params)

    async def create_session(self) -> Session:
        """Create a new session on this target."""
        response = await self._broker._browser.send_command(  # noqa: SLF001 yeah we need the browser :-(
            "Target.attachToTarget",
            params={"targetId": self.target_id, "flatten": True},
        )
        if "error" in response:
            raise RuntimeError(
                "Could not create session",
            ) from protocol.DevtoolsProtocolError(
                response,
            )
        session_id = response["result"]["sessionId"]
        new_session = Session(session_id, self._broker)
        self._add_session(new_session)
        return new_session

    # async only
    async def close_session(
        self,
        session_id: str,
    ) -> protocol.BrowserResponse:
        """
        Close a session by session_id.

        Args:
            session_id: the session to close

        """
        if isinstance(session_id, Session):
            session_id = session_id.session_id
        response = await self._broker._browser.send_command(  # noqa: SLF001 we need browser
            command="Target.detachFromTarget",
            params={"sessionId": session_id},
        )

        self._remove_session(session_id)
        if "error" in response:
            raise RuntimeError(
                "Could not close session",
            ) from protocol.DevtoolsProtocolError(
                response,
            )
        _logger.debug(f"The session {session_id} has been closed.")
        return response
        # kinda hate, why do we need this again?

    def subscribe(
        self,
        string: str,
        callback: Callable[[protocol.BrowserResponse], Coroutine[Any, Any, Any]],
        *,
        repeating: bool = True,
    ) -> None:
        """
        Subscribe to an event on the main session of this target.

        Args:
            string: the name of the event. Can use * wildcard at the end.
            callback: the callback (which takes a message dict and returns nothing)
            repeating: default True, should the callback execute more than once

        """
        session = self.get_session()
        session.subscribe(string, callback, repeating=repeating)

    def unsubscribe(self, string: str) -> None:
        """
        Remove a subscription.

        Args:
            string: the subscription to remove.

        """
        session = self.get_session()
        session.unsubscribe(string)

    def subscribe_once(self, string: str) -> asyncio.Future[Any]:
        """
        Return a future for a browser event for the first session of this target.

        Generally python asyncio doesn't recommend futures.

        But in this case, one must call subscribe_once and await it later,
        generally because they must subscribe and then provoke the event.

        Args:
            string: the event to subscribe to

        Returns:
            A future to be awaited later, the complete event.

        """
        session = self.get_session()
        return session.subscribe_once(string)



import asyncio
import os
import platform
import signal
import subprocess

import logistro
import pytest
from async_timeout import timeout

import choreographer as choreo
from choreographer import errors

# ruff: noqa: PLR0913 (lots of parameters)

# allows to create a browser pool for tests
pytestmark = pytest.mark.asyncio(loop_scope="function")

_logger = logistro.getLogger(__name__)


@pytest.mark.asyncio(loop_scope="function")
async def test_context(headless, sandbox, gpu):
    _logger.info("testing...")
    async with timeout(pytest.default_timeout):
        async with choreo.Browser(
            headless=headless,
            enable_sandbox=sandbox,
            enable_gpu=gpu,
        ) as browser:
            if sandbox and "ubuntu" in platform.version().lower():
                pytest.skip(
                    "Ubuntu doesn't support sandbox unless installed from snap.",
                )
            elif sandbox:
                _logger.info(
                    "Not skipping sandbox: "
                    f"Sandbox: {sandbox},"
                    f"Version: {platform.version().lower()}",
                )
            response = await browser.send_command(command="Target.getTargets")
            assert "result" in response and "targetInfos" in response["result"]  # noqa: PT018 combined assert
            if len(response["result"]["targetInfos"]) == 0:
                await browser.create_tab()
            assert isinstance(browser.get_tab(), choreo.Tab)
            assert len(browser.get_tab().sessions) == 1
        # let asyncio do some cleaning up if it wants, may prevent warnings
        await asyncio.sleep(0)
        assert not browser._browser_impl.tmp_dir.exists  # noqa: SLF001


@pytest.mark.asyncio(loop_scope="function")
async def test_no_context(headless, sandbox, gpu):
    _logger.info("testing...")
    browser = await choreo.Browser(
        headless=headless,
        enable_sandbox=sandbox,
        enable_gpu=gpu,
    )
    if sandbox and "ubuntu" in platform.version().lower():
        pytest.skip("Ubuntu doesn't support sandbox unless installed from snap.")
    try:
        async with timeout(pytest.default_timeout):
            response = await browser.send_command(command="Target.getTargets")
            assert "result" in response and "targetInfos" in response["result"]  # noqa: PT018 combined assert
            if len(response["result"]["targetInfos"]) == 0:
                await browser.create_tab()
            assert isinstance(browser.get_tab(), choreo.Tab)
            assert len(browser.get_tab().sessions) == 1
    finally:
        await browser.close()
    await asyncio.sleep(0)
    assert not browser._browser_impl.tmp_dir.exists  # noqa: SLF001


# Harass choreographer with a kill in this test to see if its clean in a way
# tempdir may survive protected by chromium subprocess surviving the kill
@pytest.mark.asyncio(loop_scope="function")
async def test_watchdog(headless):
    _logger.info("testing...")
    browser = await choreo.Browser(
        headless=headless,
    )

    if platform.system() == "Windows":
        # Blocking process here because it ensures the kill will occur rn
        subprocess.call(  # noqa: S603, ASYNC221 sanitize input, blocking process
            ["taskkill", "/F", "/T", "/PID", str(browser.subprocess.pid)],  # noqa: S607
            stderr=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
        )
    else:
        os.kill(browser.subprocess.pid, signal.SIGKILL)
    await asyncio.sleep(1.5)

    with pytest.raises(
        (errors.ChannelClosedError, errors.BrowserClosedError),
    ):
        await browser.send_command(command="Target.getTargets")

    await browser.close()
    await asyncio.sleep(0)



"""Contains functions and class that primarily help us with the OS."""

from ._tmpfile import TmpDirectory, TmpDirWarning
from ._which import get_browser_path

__all__ = [
    "TmpDirWarning",
    "TmpDirectory",
    "get_browser_path",
]



import argparse
import asyncio
import platform
import subprocess
import sys
import time

# diagnose function is too weird and ruff guts it
# ruff has line-level and file-level QA suppression
# so lets give diagnose a separate file

# ruff: noqa: PLR0915, C901, S603, BLE001, S607, PERF203, TRY002, T201, PLR0912, SLF001
# ruff: noqa: F401, ERA001 # temporary, sync

# in order, exceptions are:
# - function complexity (statements?)
# - function complexity (algo measure)
# - validate subprocess input arguments
# - blind exception
# - partial executable path (bash not /bin/bash)
# - performance overhead of try-except in loop
# - make own exceptions
# - no print


def diagnose() -> None:
    from choreographer import Browser, BrowserSync
    from choreographer.browsers._chrome_constants import chrome_names
    from choreographer.utils._which import browser_which

    parser = argparse.ArgumentParser(description="tool to help debug problems")
    parser.add_argument("--no-run", dest="run", action="store_false")
    parser.add_argument("--show", dest="headless", action="store_false")
    parser.set_defaults(run=True)
    parser.set_defaults(headless=True)
    args, _ = parser.parse_known_args()
    run = args.run
    headless = args.headless
    fail = []
    print("*".center(50, "*"))
    print("SYSTEM:".center(50, "*"))
    print(platform.system())
    print(platform.release())
    print(platform.version())
    print(platform.uname())
    print("BROWSER:".center(50, "*"))
    browser_path = browser_which(chrome_names)
    print(browser_path)
    print("BROWSER_INIT_CHECK (DEPS)".center(50, "*"))
    if not browser_path:
        print("No browser, found can't check for deps.")
    else:
        b = Browser()
        b._browser_impl.pre_open()
        cli = b._browser_impl.get_cli()
        env = b._browser_impl.get_env()
        b._browser_impl.clean()
        del b
        print("cli:")
        for arg in cli:
            print(arg)
        print("env:")
        for k, v in env.items():
            print(f"{k}:{v}")
    print("VERSION INFO:".center(50, "*"))
    try:
        print("PIP:".center(25, "*"))
        print(subprocess.check_output([sys.executable, "-m", "pip", "freeze"]).decode())
    except BaseException as e:
        print(f"Error w/ pip: {e}")
    try:
        print("UV:".center(25, "*"))
        print(subprocess.check_output(["uv", "pip", "freeze"]).decode())
    except BaseException as e:
        print(f"Error w/ uv: {e}")
    try:
        print("GIT:".center(25, "*"))
        print(
            subprocess.check_output(
                ["git", "describe", "--tags", "--long", "--always"],
            ).decode(),
        )
    except BaseException as e:
        print(f"Error w/ git: {e}")
    finally:
        print(sys.version)
        print(sys.version_info)
        print("Done with version info.".center(50, "*"))
    if run:
        try:
            print("Skipping sync test...")
            # print("Sync Test Headless".center(50, "*"))
            # browser = BrowserSync(headless=headless)
            # browser.open()
            # time.sleep(3)
            # browser.close()
        except BaseException as e:
            fail.append(("Sync test headless", e))
        finally:
            print("Done with sync test headless".center(50, "*"))

        async def test_headless() -> None:
            browser = await Browser(headless=headless)
            await asyncio.sleep(3)
            await browser.close()

        try:
            print("Async Test Headless".center(50, "*"))
            asyncio.run(test_headless())
        except BaseException as e:
            fail.append(("Async test headless", e))
        finally:
            print("Done with async test headless".center(50, "*"))
    print()
    sys.stdout.flush()
    sys.stderr.flush()
    if fail:
        import traceback

        for exception in fail:
            try:
                print(f"Error in: {exception[0]}")
                traceback.print_exception(
                    type(exception[1]),
                    exception[1],
                    exception[1].__traceback__,
                )
            except BaseException:
                print("Couldn't print traceback for:")
                print(str(exception))
        raise BaseException("There was an exception, see above.")
    print("Thank you! Please share these results with us!")



import asyncio

import logistro
import pytest

# allows to create a browser pool for tests
pytestmark = pytest.mark.asyncio(loop_scope="function")

_logger = logistro.getLogger(__name__)


async def test_placeholder(browser):
    _logger.info("testing...")
    assert "result" in await browser.send_command("Target.getTargets")
    await asyncio.sleep(0)



"""
Provides various implementations of Session and Target.

It includes helpers and constants for the Chrome Devtools Protocol.
"""

from __future__ import annotations

from enum import Enum
from typing import Any, MutableMapping, NewType, Optional, Tuple, cast

BrowserResponse = NewType("BrowserResponse", MutableMapping[str, Any])
"""The type for a response from the browser. Is really a `dict()`."""
BrowserCommand = NewType("BrowserCommand", MutableMapping[str, Any])
"""The type for a command to the browser. Is really a `dict()`."""

MessageKey = NewType("MessageKey", Tuple[str, Optional[int]])
"""The type for id'ing a message/response. It is `tuple(session_id, message_id)`."""


class Ecode(Enum):
    """Ecodes are a list of possible error codes chrome returns."""

    TARGET_NOT_FOUND = -32602
    """Self explanatory."""


class DevtoolsProtocolError(Exception):
    """Raise a general error reported by the devtools protocol."""

    def __init__(self, response: BrowserResponse) -> None:
        """
        Construct a new DevtoolsProtocolError.

        Args:
            response: the json response that contains the error

        """
        super().__init__(response)
        self.code = response["error"]["code"]
        self.message = response["error"]["message"]


class MessageTypeError(TypeError):
    """An error for poorly formatted devtools protocol message."""

    def __init__(self, key: str, value: Any, expected_type: type) -> None:
        """
        Construct a message about a poorly formed protocol message.

        Args:
            key: the key that has the badly typed value
            value: the type of the value that is incorrect
            expected_type: the type that was expected

        """
        value = type(value) if not isinstance(value, type) else value
        super().__init__(
            f"Message with key {key} must have type {expected_type}, not {value}.",
        )


class MissingKeyError(ValueError):
    """An error for poorly formatted devtools protocol message."""

    def __init__(self, key: str, obj: BrowserCommand) -> None:
        """
        Construct a MissingKeyError specifying which key was missing.

        Args:
            key: the missing key
            obj: the message without the key

        """
        super().__init__(
            f"Message missing required key/s {key}. Message received: {obj}",
        )


class ExperimentalFeatureWarning(UserWarning):
    """An warning to report that a feature may or may not work."""


def verify_params(obj: BrowserCommand) -> None:
    """
    Verify the message obj hast he proper keys and values.

    Args:
        obj: the object to check.

    Raises:
        MissingKeyError: if a key is missing.
        MessageTypeError: if a value type is incorrect.
        RuntimeError: if there are strange keys.

    """
    n_keys = 0

    required_keys = {"id": int, "method": str}
    for key, type_key in required_keys.items():
        if key not in obj:
            raise MissingKeyError(key, obj)
        if not isinstance(obj[key], type_key):
            raise MessageTypeError(key, type(obj[key]), type_key)
    n_keys += 2

    if "params" in obj:
        n_keys += 1
    if "sessionId" in obj:
        n_keys += 1

    if len(obj.keys()) != n_keys:
        raise RuntimeError(
            "Message objects must have id and method keys, "
            "and may have params and sessionId keys.",
        )


def calculate_message_key(msg: BrowserResponse | BrowserCommand) -> MessageKey | None:
    """
    Given a message to/from the browser, calculate the key corresponding to the command.

    Every message is uniquely identified by its sessionId and id (counter).

    Args:
        msg: the message for which to calculate the key.

    """
    session_id = msg.get("sessionId", "")
    message_id = msg.get("id")
    if message_id is None:
        return None
    return MessageKey((session_id, message_id))


def match_message_key(response: BrowserResponse, key: MessageKey) -> bool:
    """
    Report True if a response matches with a certain key (sessionId, id).

    Args:
        response: the object response from the browser
        key: the (sessionId, id) key tubple we're looking for

    """
    session_id, message_id = key
    if ("session_id" not in response and session_id == "") or (  # is browser session
        "session_id" in response and response["session_id"] == session_id  # is session
    ):
        pass
    else:
        return False

    if "id" in response and str(response["id"]) == str(message_id):
        pass
    else:
        return False
    return True


def is_event(response: BrowserResponse) -> bool:
    """Return true if the browser response is an event notification."""
    required_keys = {"method", "params"}
    return required_keys <= response.keys() and "id" not in response


def get_target_id_from_result(response: BrowserResponse) -> str | None:
    """
    Extract target id from a browser response.

    Args:
        response: the browser response to extract the targetId from.

    """
    if "result" in response and "targetId" in response["result"]:
        return cast("str", response["result"]["targetId"])
    else:
        return None


def get_session_id_from_result(response: BrowserResponse) -> str | None:
    """
    Extract session id from a browser response.

    Args:
        response: the browser response to extract the sessionId from.

    """
    if "result" in response and "sessionId" in response["result"]:
        return cast("str", response["result"]["sessionId"])
    else:
        return None


def get_error_from_result(response: BrowserResponse) -> str | None:
    """
    Extract error from a browser response.

    Args:
        response: the browser response to extract the error from.

    """
    if "error" in response:
        return cast("str", response["error"])
    else:
        return None



from __future__ import annotations

import argparse
import asyncio
import json
import platform
import shutil
import sys
import urllib.request
import warnings
import zipfile
from functools import partial
from pathlib import Path

import logistro

from choreographer.cli.defaults import default_download_path

_logger = logistro.getLogger(__name__)

_chrome_for_testing_url = "https://googlechromelabs.github.io/chrome-for-testing/known-good-versions-with-downloads.json"


_platforms = ["linux64", "win32", "win64", "mac-x64", "mac-arm64"]

_arch_size_detected = "64" if sys.maxsize > 2**32 else "32"
_arch_detected = "arm" if platform.processor() == "arm" else "x"

if platform.system() == "Windows":
    _chrome_platform_detected = "win" + _arch_size_detected
elif platform.system() == "Linux":
    _chrome_platform_detected = "linux" + _arch_size_detected
elif platform.system() == "Darwin":
    _chrome_platform_detected = "mac-" + _arch_detected + _arch_size_detected

_default_exe_path = Path()
if platform.system().startswith("Linux"):
    _default_exe_path = (
        default_download_path / f"chrome-{_chrome_platform_detected}" / "chrome"
    )
elif platform.system().startswith("Darwin"):
    _default_exe_path = (
        default_download_path
        / f"chrome-{_chrome_platform_detected}"
        / "Google Chrome for Testing.app"
        / "Contents"
        / "MacOS"
        / "Google Chrome for Testing"
    )
elif platform.system().startswith("Win"):
    _default_exe_path = (
        default_download_path / f"chrome-{_chrome_platform_detected}" / "chrome.exe"
    )


def get_chrome_download_path() -> Path:
    return _default_exe_path


# https://stackoverflow.com/questions/39296101/python-zipfile-removes-execute-permissions-from-binaries
class _ZipFilePermissions(zipfile.ZipFile):
    def _extract_member(self, member, targetpath, pwd):  # type: ignore [no-untyped-def]
        if not isinstance(member, zipfile.ZipInfo):
            member = self.getinfo(member)

        path = super()._extract_member(member, targetpath, pwd)  # type: ignore [misc]
        # High 16 bits are os specific (bottom is st_mode flag)
        attr = member.external_attr >> 16
        if attr != 0:
            Path(path).chmod(attr)
        return path


def get_chrome_sync(
    arch: str = _chrome_platform_detected,
    i: int | None = None,
    path: str | Path = default_download_path,
    *,
    verbose: bool = False,
) -> Path | str:
    """Download chrome synchronously: see `get_chrome()`."""
    if isinstance(path, str):
        path = Path(path)
    if i:
        _logger.info("Loading chrome from list")
        browser_list = json.loads(
            urllib.request.urlopen(  # noqa: S310 audit url for schemes
                _chrome_for_testing_url,
            ).read(),
        )
        version_obj = browser_list["versions"][i]
    else:
        _logger.info("Using last known good version of chrome")
        with (
            Path(__file__).resolve().parent.parent
            / "resources"
            / "last_known_good_chrome.json"
        ).open() as f:
            version_obj = json.load(f)
    if verbose:
        print(version_obj["version"])  # noqa: T201 allow print in cli
        print(version_obj["revision"])  # noqa: T201 allow print in cli
    chromium_sources = version_obj["downloads"]["chrome"]
    url = ""
    for src in chromium_sources:
        if src["platform"] == arch:
            url = src["url"]
            break
    if not path.exists():
        path.mkdir(parents=True)
    filename = path / "chrome.zip"
    with urllib.request.urlopen(url) as response, filename.open("wb") as out_file:  # noqa: S310 audit url
        shutil.copyfileobj(response, out_file)
    with _ZipFilePermissions(filename, "r") as zip_ref:
        zip_ref.extractall(path)

    if arch.startswith("linux"):
        exe_name = path / f"chrome-{arch}" / "chrome"
    elif arch.startswith("mac"):
        exe_name = (
            path
            / f"chrome-{arch}"
            / "Google Chrome for Testing.app"
            / "Contents"
            / "MacOS"
            / "Google Chrome for Testing"
        )
    elif arch.startswith("win"):
        exe_name = path / f"chrome-{arch}" / "chrome.exe"

    return exe_name


async def get_chrome(
    arch: str = _chrome_platform_detected,
    i: int | None = None,
    path: str | Path = default_download_path,
    *,
    verbose: bool = False,
) -> Path | str:
    """
    Download google chrome from google-chrome-for-testing server.

    Args:
        arch: the target platform/os, as understood by google's json directory.
        i: the chrome version: -1 being the latest version, 0 being the oldest
           still in the testing directory.
        path: where to download it too (the folder).
        verbose: print out version found

    """
    loop = asyncio.get_running_loop()
    fn = partial(get_chrome_sync, arch=arch, i=i, path=path, verbose=verbose)
    return await loop.run_in_executor(
        executor=None,
        func=fn,
    )


def get_chrome_cli() -> None:
    if "ubuntu" in platform.version().lower():
        warnings.warn(  # noqa: B028
            "You are using `get_browser()` on Ubuntu."
            " Ubuntu is **very strict** about where binaries come from."
            " While sandbox is already off by default, do not set"
            " enable_sandbox to True OR you can install from Ubuntu's"
            " package manager.",
            UserWarning,
        )
    parser = argparse.ArgumentParser(
        description="Will download Chrome for testing. All arguments optional.",
        parents=[logistro.parser],
    )
    parser.add_argument(
        "--i",
        "-i",
        type=int,
        dest="i",
        help=(
            "Google offers thousands of chrome versions for download. "
            "'-i 0' is the oldest, '-i -1' is the newest: array syntax"
        ),
    )
    parser.add_argument(
        "--arch",
        dest="arch",
        help="linux64|win32|win64|mac-x64|mac-arm64",
    )
    parser.add_argument(
        "--path",
        dest="path",
        help="Where to store the download.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        help="Display found version number if using -i (to stdout)",
    )
    parser.set_defaults(path=default_download_path)
    parser.set_defaults(arch=_chrome_platform_detected)
    parser.set_defaults(verbose=False)
    parsed = parser.parse_args()
    i = parsed.i
    arch = parsed.arch
    path = Path(parsed.path)
    verbose = parsed.verbose
    if not arch or arch not in _platforms:
        raise RuntimeError(
            "You must specify a platform: "
            f"linux64, win32, win64, mac-x64, mac-arm64, not {platform}",
        )
    print(get_chrome_sync(arch=arch, i=i, path=path, verbose=verbose))  # noqa: T201 allow print in cli



"""Defaults used when arguments not supplied."""

from pathlib import Path

default_download_path = Path(__file__).resolve().parent / "browser_exe"
"""The path where we download chrome if no path argument is supplied."""



try:
    from datetime import UTC, datetime  #  type: ignore [attr-defined]
except ImportError:
    from datetime import datetime, timezone

    UTC = timezone.utc

import logistro
import numpy as np
import pytest

import choreographer.channels._wire as wire

# allows to create a browser pool for tests
pytestmark = pytest.mark.asyncio(loop_scope="function")


_timestamp = datetime(1970, 1, 1, tzinfo=UTC)

data = [1, 2.00, 3, float("nan"), float("inf"), float("-inf"), _timestamp]
expected_message = b'[1, 2.0, 3, null, null, null, "1970-01-01T00:00:00+00:00"]'
converted_type = [int, float, int, type(None), type(None), type(None), str]

_logger = logistro.getLogger(__name__)


@pytest.mark.asyncio
async def test_de_serialize():
    _logger.info("testing...")
    message = wire.serialize(data)
    assert message == expected_message
    obj = wire.deserialize(message)
    assert len(obj) == len(converted_type)
    for o, t in zip(obj, converted_type):
        assert isinstance(o, t)
    message_np = wire.serialize(np.array(data))
    assert message_np == expected_message
    obj_np = wire.deserialize(message_np)
    assert len(obj_np) == len(converted_type)
    for o, t in zip(obj_np, converted_type):
        assert isinstance(o, t)



import logistro
import pytest

import choreographer as choreo
from choreographer import errors
from choreographer.protocol import devtools_async

# We no longer use live URLs to as not depend on the network

# allows to create a browser pool for tests
pytestmark = pytest.mark.asyncio(loop_scope="function")

_logger = logistro.getLogger(__name__)


@pytest.mark.asyncio
async def test_create_and_close_tab(browser):
    _logger.info("testing...")
    tab = await browser.create_tab("")
    assert isinstance(tab, choreo.Tab)
    assert tab.target_id in browser.tabs
    await browser.close_tab(tab)
    assert tab.target_id not in browser.tabs


@pytest.mark.asyncio
async def test_create_and_close_session(browser):
    _logger.info("testing...")
    with pytest.warns(errors.ExperimentalFeatureWarning):
        session = await browser.create_session()
    assert isinstance(session, devtools_async.Session)
    assert session.session_id in browser.sessions
    await browser.close_session(session)
    assert session.session_id not in browser.sessions


# Along with testing, this could be repurposed as well to diagnose
# This deserves some thought re. difference between write_json and send_command
@pytest.mark.asyncio
async def test_broker_write_json(browser):
    _logger.info("testing...")
    # Test valid request with correct id and method
    response = await browser._broker.write_json(  # noqa: SLF001
        {"id": 0, "method": "Target.getTargets"},
    )
    assert "result" in response and "targetInfos" in response["result"]  # noqa: PT018 I like this assertion

    # Test invalid method name should return error
    response = await browser._broker.write_json(  # noqa: SLF001
        {"id": 2, "method": "dkadklqwmd"},
    )
    assert "error" in response

    # Test missing 'id' key
    with pytest.raises(
        errors.MissingKeyError,
    ):
        await browser._broker.write_json(  # noqa: SLF001
            {"method": "Target.getTargets"},
        )

    # Test missing 'method' key
    with pytest.raises(
        errors.MissingKeyError,
    ):
        await browser._broker.write_json(  # noqa: SLF001
            {"id": 1},
        )

    # Test empty dictionary
    with pytest.raises(
        errors.MissingKeyError,
    ):
        await browser._broker.write_json({})  # noqa: SLF001

    # Test invalid parameter in the message
    with pytest.raises(
        RuntimeError,
    ):
        await browser._broker.write_json(  # noqa: SLF001
            {"id": 0, "method": "Target.getTargets", "invalid_parameter": "kamksamdk"},
        )

    # Test int method should return error
    with pytest.raises(
        errors.MessageTypeError,
    ):
        await browser._broker.write_json(  # noqa: SLF001
            {"id": 3, "method": 12345},
        )

    # Test non-integer id should return error
    with pytest.raises(
        errors.MessageTypeError,
    ):
        await browser._broker.write_json(  # noqa: SLF001
            {"id": "string", "method": "Target.getTargets"},
        )


@pytest.mark.asyncio
async def test_browser_send_command(browser):
    _logger.info("testing...")
    # Test valid request with correct command
    response = await browser.send_command(command="Target.getTargets")
    assert "result" in response and "targetInfos" in response["result"]  # noqa: PT018 I like this assertion

    # Test invalid method name should return error
    response = await browser.send_command(command="dkadklqwmd")
    assert "error" in response

    # Test int method should return error
    with pytest.raises(
        errors.MessageTypeError,
    ):
        await browser.send_command(command=12345)


@pytest.mark.asyncio
async def test_populate_targets(browser):
    _logger.info("testing...")
    await browser.send_command(command="Target.createTarget", params={"url": ""})
    await browser.populate_targets()
    assert len(browser.tabs) >= 1


@pytest.mark.asyncio
async def test_get_tab(browser):
    _logger.info("testing...")
    await browser.create_tab("")
    assert browser.get_tab() == next(iter(browser.tabs.values()))
    await browser.create_tab()
    await browser.create_tab("")
    assert browser.get_tab() == next(iter(browser.tabs.values()))



"""
_unix_pipe_chromium_wrapper.py provides proper fds to chrome.

By running chromium in a new process (this wrapper), we guarantee
the user hasn't stolen one of our desired file descriptors, which
the OS gives away first-come-first-serve everytime someone opens a
file. chromium demands we use 3 and 4.
"""

from __future__ import annotations

import os

# importing modules has side effects, so we do this before imports
# ruff: noqa: E402

# chromium reads on 3, writes on 4
os.dup2(0, 3)  # make our stdin their input
os.dup2(1, 4)  # make our stdout their output

_inheritable = True
os.set_inheritable(4, _inheritable)
os.set_inheritable(3, _inheritable)

import signal
import subprocess
import sys
from functools import partial
from typing import TYPE_CHECKING

import logistro

if TYPE_CHECKING:
    from types import FrameType

_logger = logistro.getLogger("chrome_wrapper")

# we're a wrapper, the cli is everything that came after us
cli = sys.argv[1:]

print(f"wrapper CLI: {cli}", file=sys.stderr)  # noqa: T201 goes to pipe/logger anyway

process = subprocess.Popen(cli, pass_fds=(3, 4))  # noqa: S603 untrusted input


def kill_proc(
    process: subprocess.Popen[bytes],
    _sig_num: int,
    _frame: FrameType | None,
) -> None:
    process.terminate()
    process.wait(5)  # 5 seconds to clean up nicely, it's a lot
    process.kill()


kp = partial(kill_proc, process)
signal.signal(signal.SIGTERM, kp)
signal.signal(signal.SIGINT, kp)

process.wait()

# not great but it seems that
# pipe isn't always closed when chrome closes
# so we pretend to be chrome and send a bye instead
# also, above depends on async/sync, platform, etc
print("{bye}")  # noqa: T201 we need print here



"""Provides a channel based on operating system file pipes."""

from __future__ import annotations

import os
import platform
import sys
import warnings
from threading import Lock
from typing import TYPE_CHECKING

import logistro

from . import _wire as wire
from ._errors import BlockWarning, ChannelClosedError, JSONError

if TYPE_CHECKING:
    from typing import Any, Mapping, Sequence

    from choreographer.protocol import BrowserResponse

_with_block = bool(sys.version_info[:3] >= (3, 12) or platform.system() != "Windows")

_logger = logistro.getLogger(__name__)


# if we're a pipe we expect these public attributes
class Pipe:
    """Defines an operating system pipe."""

    from_external_to_choreo: int
    """Consumers need this, it is the channel the browser uses to talk to choreo."""
    from_choreo_to_external: int
    """Consumers needs this, it is the channel choreo writes to the browser on."""
    shutdown_lock: Lock
    """Once this is locked, the pipe is closed and can't be reopened."""

    def __init__(self) -> None:
        """Construct a pipe using os functions."""
        # This is where pipe listens (from browser)
        # So pass the write to browser
        self._read_from_browser, self._write_from_browser = list(os.pipe())

        # This is where pipe writes (to browser)
        # So pass the read to browser
        self._read_to_browser, self._write_to_browser = list(os.pipe())

        # Popen will write stdout of wrapper to this (dupping 4)
        # Browser will write directly to this if not using wrapper
        self.from_external_to_choreo = self._write_from_browser
        # Popen will read this into stdin of wrapper (dupping 3)
        # Browser will read directly from this if not using wrapper
        # which dupes stdin to expected fd (4?)
        self.from_choreo_to_external = self._read_to_browser
        # These won't be used on windows directly, they'll be t-formed to
        # windows-style handles. But let another layer handle that.

        # this is just a convenience to prevent multiple shutdowns
        self.shutdown_lock = Lock()  # should be private

    def write_json(self, obj: Mapping[str, Any]) -> None:
        """
        Send one json down the pipe.

        Args:
            obj: any python object that serializes to json.

        """
        if self.shutdown_lock.locked():
            raise ChannelClosedError
        encoded_message = wire.serialize(obj) + b"\0"
        _logger.debug(
            f"Writing message {encoded_message[:15]!r}...{encoded_message[-15:]!r}, "
            f"size: {len(encoded_message)}.",
        )
        _logger.debug2(f"Full Message: {encoded_message!r}")
        try:
            ret = os.write(self._write_to_browser, encoded_message)
            _logger.debug(
                f"***Wrote {ret}/{len(encoded_message)}***",
            )
            if ret != len(encoded_message):
                _logger.critical(
                    f"***Did not write entire message. {ret}/{len(encoded_message)}***",
                )
        except OSError as e:
            self.close()
            raise ChannelClosedError from e

    def read_jsons(  # noqa: PLR0912, PLR0915, C901 branches, complexity
        self,
        *,
        blocking: bool = True,
    ) -> Sequence[BrowserResponse]:
        """
        Read from the pipe and return one or more jsons in a list.

        Args:
            blocking: The read option can be set to block or not.

        Returns:
            A list of jsons.

        """
        if self.shutdown_lock.locked():
            raise ChannelClosedError
        if not _with_block and not blocking:
            warnings.warn(  # noqa: B028
                "Windows python version < 3.12 does not support non-blocking",
                BlockWarning,
            )
        jsons: list[BrowserResponse] = []
        try:
            if _with_block:
                os.set_blocking(self._read_from_browser, blocking)
        except OSError as e:
            self.close()
            raise ChannelClosedError from e
        try:
            loop_count = 1
            raw_buffer = None  # if we fail in read, we already defined
            raw_buffer = os.read(
                self._read_from_browser,
                10000,
            )  # 10MB buffer, nbd, doesn't matter w/ this
            _logger.debug(
                f"First read in loop: {raw_buffer[:15]!r}...{raw_buffer[-15:]!r}. "
                f"size: {len(raw_buffer)}.",
            )
            _logger.debug2(f"Whole buffer: {raw_buffer!r}")
            if not raw_buffer or raw_buffer == b"{bye}\n":
                if raw_buffer:
                    _logger.debug(f"Received {raw_buffer!r}. is bye?")
                # we seem to need {bye} even if chrome closes NOTE
                self.close()
                raise ChannelClosedError
            while raw_buffer[-1] != 0:
                _logger.debug("Partial message from browser received.")
                loop_count += 1
                if _with_block:
                    os.set_blocking(self._read_from_browser, True)
                raw_buffer += os.read(self._read_from_browser, 10000)
        except BlockingIOError:
            _logger.debug("BlockingIOError")
            return jsons
        except OSError as e:
            _logger.debug("OSError")
            self.close()
            if not raw_buffer or raw_buffer == b"{bye}\n":
                raise ChannelClosedError from e
            # this could be hard to test as it is a real OS corner case
        finally:
            _logger.debug(
                f"Total loops: {loop_count}, "
                f"Final size: {len(raw_buffer) if raw_buffer else 0}.",
            )
            _logger.debug2(f"Whole buffer: {raw_buffer!r}")
        decoded_buffer = raw_buffer.decode("utf-8")
        raw_messages = decoded_buffer.split("\0")
        _logger.debug(f"Received {len(raw_messages)} raw_messages.")
        for raw_message in raw_messages:
            if raw_message:
                try:
                    jsons.append(wire.deserialize(raw_message))
                except JSONError:
                    _logger.exception("JSONError decoding message. Ignoring")
                except:
                    _logger.exception("Error in trying to decode JSON off our read.")
                    raise
        return jsons

    def _unblock_fd(self, fd: int) -> None:
        try:
            if _with_block:
                os.set_blocking(fd, False)
        except BaseException:  # noqa: BLE001, S110 OS errors are not consistent, catch blind + pass
            pass

    def _close_fd(self, fd: int) -> None:
        try:
            os.close(fd)
        except BaseException:  # noqa: BLE001, S110 OS errors are not consistent, catch blind + pass
            pass

    def _fake_bye(self) -> None:
        self._unblock_fd(self._write_from_browser)
        try:
            os.write(self._write_from_browser, b"{bye}\n")
        except BaseException:  # noqa: BLE001, S110 OS errors are not consistent, catch blind + pass
            pass

    def close(self) -> None:
        """Close the pipe."""
        if self.shutdown_lock.acquire(blocking=False):
            if platform.system() == "Windows":
                self._fake_bye()
            self._unblock_fd(self._write_from_browser)
            self._unblock_fd(self._read_from_browser)
            self._unblock_fd(self._write_to_browser)
            self._unblock_fd(self._read_to_browser)
            self._close_fd(self._write_to_browser)  # no more writes
            self._close_fd(self._write_from_browser)  # we're done with writes
            self._close_fd(self._read_from_browser)  # no more attempts at read
            self._close_fd(self._read_to_browser)



class BlockWarning(UserWarning):
    """A warning for when block modification operatins used on incompatible OS."""


class ChannelClosedError(IOError):
    """An error to throw when the channel has closed from either end or error."""


class JSONError(RuntimeError):
    """Another JSONError."""



"""cli provides some tools that are used on the commandline (and to download chrome)."""

from ._cli_utils import (
    get_chrome,
    get_chrome_sync,
)

__all__ = [
    "get_chrome",
    "get_chrome_sync",
]



"""Provides the sync api: `BrowserSync`, `TabSync`."""

from __future__ import annotations

import os
import subprocess
from threading import Lock
from typing import TYPE_CHECKING

import logistro

from ._brokers import BrokerSync
from .browsers import BrowserClosedError, BrowserFailedError, Chromium
from .channels import ChannelClosedError, Pipe
from .protocol.devtools_sync import SessionSync, TargetSync
from .utils._kill import kill

if TYPE_CHECKING:
    from pathlib import Path
    from types import TracebackType
    from typing import Any, MutableMapping

    from typing_extensions import Self  # 3.9 needs this, could be from typing in 3.10

    from .browsers._interface_type import BrowserImplInterface
    from .channels._interface_type import ChannelInterface

_logger = logistro.getLogger(__name__)


class TabSync(TargetSync):
    """A wrapper for `TargetSync`, so user can use `TabSync`, not `TargetSync`."""


class BrowserSync(TargetSync):
    """`BrowserSync` is the sync implementation of `Browser`."""

    # A list of the types that are essential to use
    # with this class

    tabs: MutableMapping[str, TabSync]
    """A mapping by target_id of all the targets which are open tabs."""
    targets: MutableMapping[str, TargetSync]
    """A mapping by target_id of ALL the targets."""
    # Don't init instance attributes with mutables

    def _make_lock(self) -> None:
        self._open_lock = Lock()

    def _is_open(self) -> bool:
        return not self._open_lock.acquire(blocking=False)

    def _release_lock(self) -> bool:
        try:
            if self._open_lock.locked():
                self._open_lock.release()
                return True
            else:
                return False
        except RuntimeError:
            return False

    def __init__(
        self,
        path: str | Path | None = None,
        *,
        browser_cls: type[BrowserImplInterface] = Chromium,
        channel_cls: type[ChannelInterface] = Pipe,
        **kwargs: Any,
    ) -> None:
        """
        Construct a new browser instance.

        Args:
            path: The path to the browser executable.
            browser_cls: The type of browser (default: `Chromium`).
            channel_cls: The type of channel to browser (default: `Pipe`).
            kwargs: The arguments that the browser_cls takes. For example,
                headless=True/False, enable_gpu=True/False, etc.

        """
        _logger.debug("Attempting to open new browser.")
        self._make_lock()
        self.tabs = {}
        self.targets = {}

        # Compose Resources
        self._channel = channel_cls()
        self._broker = BrokerSync(self, self._channel)
        self._browser_impl = browser_cls(self._channel, path, **kwargs)
        if hasattr(browser_cls, "logger_parser"):
            parser = browser_cls.logger_parser
        else:
            parser = None
        self._logger_pipe, _ = logistro.getPipeLogger(
            "browser_proc",
            parser=parser,
        )

    def open(self) -> None:
        """Open the browser."""
        if self._is_open():
            raise RuntimeError("Can't re-open the browser")
        self.subprocess = subprocess.Popen(  # noqa: S603
            self._browser_impl.get_cli(),
            stderr=self._logger_pipe,
            env=self._browser_impl.get_env(),
            **self._browser_impl.get_popen_args(),
        )
        super().__init__("0", self._broker)
        self._add_session(SessionSync("", self._broker))

    def __enter__(self) -> Self:
        """Open browser as context to launch on entry and close on exit."""
        self.open()
        return self

    def _is_closed(self, wait: int | None = 0) -> bool:
        if wait == 0:
            return self.subprocess.poll() is None
        else:
            try:
                self.subprocess.wait(wait)
            except subprocess.TimeoutExpired:
                return False
        return True

    def _close(self) -> None:
        if self._is_closed():
            return

        try:
            self.send_command("Browser.close")
        except (BrowserClosedError, BrowserFailedError):
            _logger.debug("Browser is closed trying to send Browser.close")
            return
        except ChannelClosedError:
            _logger.debug("Can send browser.close on close channel")

        self._channel.close()

        if self._is_closed(wait=3):
            return

        # try kiling
        kill(self.subprocess)
        if self._is_closed(wait=4):
            return
        else:
            raise RuntimeError("Couldn't close or kill browser subprocess")

    def close(self) -> None:
        """Close the browser."""
        self._broker.clean()
        _logger.debug("Broker cleaned up.")
        if not self._release_lock():
            return
        try:
            _logger.debug("Trying to close browser.")
            self._close()
            _logger.debug("browser._close() called successfully.")
        except ProcessLookupError:
            pass
        if self._logger_pipe:
            os.close(self._logger_pipe)
        _logger.debug("Logging pipe closed.")
        self._channel.close()
        _logger.debug("Browser channel closed.")
        self._browser_impl.clean()
        _logger.debug("Browser implementation cleaned up.")

    def __exit__(
        self,
        type_: type[BaseException] | None,
        value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:  # None instead of False is fine, eases type checking
        """Close the browser."""
        self.close()

    def _add_tab(self, tab: TabSync) -> None:
        if not isinstance(tab, TabSync):
            raise TypeError("tab must be an object of TabSync")
        self.tabs[tab.target_id] = tab

    def _remove_tab(self, target_id: str) -> None:
        if isinstance(target_id, TabSync):
            target_id = target_id.target_id
        del self.tabs[target_id]

    def get_tab(self) -> TabSync | None:
        """Get the first tab if there is one. Useful for default tabs."""
        if self.tabs.values():
            return next(iter(self.tabs.values()))
        return None

    # wrap our broker for convenience
    def start_output_thread(self, **kwargs: Any) -> None:
        """
        Start a separate thread that dumps all messages received to stdout.

        Args:
            kwargs: passed directly to print().

        """
        self._broker.run_output_thread(**kwargs)



from __future__ import annotations

import asyncio
import warnings
from functools import partial
from typing import TYPE_CHECKING

import logistro

from choreographer import channels, protocol

# afrom choreographer.channels import ChannelClosedError

if TYPE_CHECKING:
    from typing import Any, MutableMapping

    from choreographer.browser_async import Browser
    from choreographer.channels._interface_type import ChannelInterface
    from choreographer.protocol.devtools_async import Session, Target


_logger = logistro.getLogger(__name__)


class UnhandledMessageWarning(UserWarning):
    pass


class Broker:
    """Broker is a middleware implementation for asynchronous implementations."""

    _browser: Browser
    """Browser is a reference to the Browser object this broker is brokering for."""
    _channel: ChannelInterface
    """
    Channel will be the ChannelInterface implementation (pipe or websocket)
    that the broker communicates on.
    """
    futures: MutableMapping[protocol.MessageKey, asyncio.Future[Any]]
    """A mapping of all the futures for all sent commands."""

    _subscriptions_futures: MutableMapping[
        str,
        MutableMapping[
            str,
            list[asyncio.Future[Any]],
        ],
    ]
    """A mapping of session id: subscription: list[futures]"""

    def __init__(self, browser: Browser, channel: ChannelInterface) -> None:
        """
        Construct a broker for a synchronous arragenment w/ both ends.

        Args:
            browser: The sync browser implementation.
            channel: The channel the browser uses to talk on.

        """
        self._browser = browser
        self._channel = channel
        self._background_tasks: set[asyncio.Task[Any]] = set()
        # if its a task you dont want canceled at close (like the close task)
        self._background_tasks_cancellable: set[asyncio.Task[Any]] = set()
        # if its a user task, can cancel
        self._current_read_task: asyncio.Task[Any] | None = None
        self.futures = {}
        self._subscriptions_futures = {}

        self._write_lock = asyncio.Lock()

    def new_subscription_future(
        self,
        session_id: str,
        subscription: str,
    ) -> asyncio.Future[Any]:
        _logger.debug(
            f"Session {session_id} is subscribing to {subscription} one time.",
        )
        if session_id not in self._subscriptions_futures:
            self._subscriptions_futures[session_id] = {}
        if subscription not in self._subscriptions_futures[session_id]:
            self._subscriptions_futures[session_id][subscription] = []
        future = asyncio.get_running_loop().create_future()
        self._subscriptions_futures[session_id][subscription].append(future)
        return future

    def clean(self) -> None:
        _logger.debug("Cancelling message futures")
        for future in self.futures.values():
            if not future.done():
                _logger.debug2(f"Cancelling {future}")
                future.cancel()
        _logger.debug("Cancelling read task")
        if self._current_read_task and not self._current_read_task.done():
            _logger.debug2(f"Cancelling read: {self._current_read_task}")
            self._current_read_task.cancel()
        _logger.debug("Cancelling subscription-futures")
        for session in self._subscriptions_futures.values():
            for query in session.values():
                for future in query:
                    if not future.done():
                        _logger.debug2(f"Cancelling {future}")
                        future.cancel()
        _logger.debug("Cancelling background tasks")
        for task in self._background_tasks_cancellable:
            if not task.done():
                _logger.debug2(f"Cancelling {task}")
                task.cancel()

    def run_read_loop(self) -> None:  # noqa: C901, PLR0915 complexity
        def check_read_loop_error(result: asyncio.Future[Any]) -> None:
            e = result.exception()
            if e:
                _logger.debug("Error in readloop. Will post a close() task.")
                self._background_tasks.add(
                    asyncio.create_task(self._browser.close()),
                )
                if isinstance(e, channels.ChannelClosedError):
                    _logger.debug("PipeClosedError caught")
                    _logger.debug2("Full Error:", exc_info=e)
                elif isinstance(e, asyncio.CancelledError):
                    _logger.debug("CancelledError caught.")
                    _logger.debug2("Full Error:", exc_info=e)
                else:
                    _logger.error("Error in run_read_loop.", exc_info=e)
                    raise e

        async def read_loop() -> None:  # noqa: PLR0912, PLR0915, C901
            loop = asyncio.get_running_loop()
            fn = partial(self._channel.read_jsons, blocking=True)
            responses = await loop.run_in_executor(
                executor=None,
                func=fn,
            )
            _logger.debug(f"Channel read found {len(responses)} json objects.")
            for response in responses:
                error = protocol.get_error_from_result(response)
                key = protocol.calculate_message_key(response)
                if not key and error:
                    raise protocol.DevtoolsProtocolError(response)

                # looks for event that we should handle internally
                self._check_for_closed_session(response)
                # surrounding lines overlap in idea
                if protocol.is_event(response):
                    event_session_id = response.get(
                        "sessionId",
                        "",
                    )
                    _logger.debug2(f"Is event for {event_session_id}")
                    x = self._get_target_session_by_session_id(
                        event_session_id,
                    )
                    if not x:
                        continue
                    _, event_session = x
                    if not event_session:
                        _logger.error("Found an event that returned no session.")
                        continue
                    _logger.debug(
                        f"Received event {response['method']} for "
                        f"{event_session_id} targeting {event_session}.",
                    )

                    session_futures = self._subscriptions_futures.get(
                        event_session_id,
                    )
                    _logger.debug2(
                        "Checking for event subscription future.",
                    )
                    if session_futures:
                        for query in session_futures:
                            match = (
                                query.endswith("*")
                                and response["method"].startswith(query[:-1])
                            ) or (response["method"] == query)
                            if match:
                                _logger.debug2(
                                    "Found event subscription future.",
                                )
                                for future in session_futures[query]:
                                    if not future.done():
                                        future.set_result(response)
                                session_futures[query] = []

                    _logger.debug2(
                        "Checking for event subscription callback.",
                    )
                    for query in list(event_session.subscriptions):
                        match = (
                            query.endswith("*")
                            and response["method"].startswith(query[:-1])
                        ) or (response["method"] == query)
                        _logger.debug2(
                            "Found event subscription callback.",
                        )
                        if match:
                            t: asyncio.Task[Any] = asyncio.create_task(
                                event_session.subscriptions[query][0](response),
                            )
                            self._background_tasks_cancellable.add(t)
                            if not event_session.subscriptions[query][1]:
                                event_session.unsubscribe(query)

                elif key:
                    _logger.debug(f"Have a response with key {key}")
                    if key in self.futures:
                        _logger.debug(f"Found future for key {key}")
                        future = self.futures.pop(key)
                    elif "error" in response:
                        raise protocol.DevtoolsProtocolError(response)
                    else:
                        raise RuntimeError(f"Couldn't find a future for key: {key}")
                    if not future.done():
                        future.set_result(response)
                else:
                    warnings.warn(
                        f"Unhandled message type:{response!s}",
                        UnhandledMessageWarning,
                        stacklevel=1,
                    )
            read_task = asyncio.create_task(read_loop())
            read_task.add_done_callback(check_read_loop_error)
            self._current_read_task = read_task

        read_task = asyncio.create_task(read_loop())
        read_task.add_done_callback(check_read_loop_error)
        self._current_read_task = read_task

    async def write_json(
        self,
        obj: protocol.BrowserCommand,
    ) -> protocol.BrowserResponse:
        protocol.verify_params(obj)
        key = protocol.calculate_message_key(obj)
        _logger.debug1(f"Broker writing {obj['method']} with key {key}")
        if not key:
            raise RuntimeError(
                "Message strangely formatted and "
                "choreographer couldn't figure it out why.",
            )
        loop = asyncio.get_running_loop()
        future: asyncio.Future[protocol.BrowserResponse] = loop.create_future()
        self.futures[key] = future
        _logger.debug(f"Created future: {key} {future}")
        try:
            async with self._write_lock:
                loop = asyncio.get_running_loop()
                await loop.run_in_executor(
                    None,
                    self._channel.write_json,
                    obj,
                )
        except BaseException as e:  # noqa: BLE001
            future.set_exception(e)
            del self.futures[key]
            _logger.debug(f"Future for {key} deleted.")
        return await future

    def _get_target_session_by_session_id(
        self,
        session_id: str,
    ) -> tuple[Target, Session] | None:
        if session_id == "":
            return (self._browser, self._browser.sessions[session_id])
        for tab in self._browser.tabs.values():
            if session_id in tab.sessions:
                return (tab, tab.sessions[session_id])
        if session_id in self._browser.sessions:
            return (self._browser, self._browser.sessions[session_id])
        return None

    def _check_for_closed_session(self, response: protocol.BrowserResponse) -> bool:
        if "method" in response and response["method"] == "Target.detachedFromTarget":
            session_closed = response["params"].get(
                "sessionId",
                "",
            )
            if session_closed == "":
                _logger.debug2("Found closed session through events.")
                return True

            x = self._get_target_session_by_session_id(session_closed)
            if x:
                target_closed, _ = x
            else:
                return False

            if target_closed:
                target_closed._remove_session(session_closed)  # noqa: SLF001
                _logger.debug(
                    "Using intern subscription key: "
                    "'Target.detachedFromTarget'. "
                    f"Session {session_closed} was closed.",
                )
                _logger.debug2("Found closed session through events.")
                return True
            return False
        else:
            return False



from __future__ import annotations

import os
import platform
import shutil
from typing import TYPE_CHECKING

import logistro

from choreographer.cli._cli_utils import get_chrome_download_path

_logger = logistro.getLogger()

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any, Sequence


def _is_exe(path: str | Path) -> bool:
    try:
        return os.access(path, os.X_OK)
    except:  # noqa: E722 bare except ok, weird errors, best effort.
        return False


def _which_from_windows_reg() -> str | None:
    try:
        import re
        import winreg

        command = winreg.QueryValueEx(  # type: ignore [attr-defined]
            winreg.OpenKey(  # type: ignore [attr-defined]
                winreg.HKEY_CLASSES_ROOT,  # type: ignore [attr-defined]
                "ChromeHTML\\shell\\open\\command",
                0,
                winreg.KEY_READ,  # type: ignore [attr-defined]
            ),
            "",
        )[0]
        exe = re.search('"(.*?)"', command).group(1)  # type: ignore [union-attr]
    except BaseException:  # noqa: BLE001 don't care why, best effort search
        return None

    return exe


def browser_which(
    executable_names: Sequence[str],
    *,
    skip_local: bool = False,
) -> str | None:
    """
    Look for and return first name found in PATH.

    Args:
        executable_names: the list of names to look for
        skip_local: (default False) don't look for a choreo download of anything.

    """
    _logger.debug(f"Looking for browser, skipping local? {skip_local}")
    path = None

    if isinstance(executable_names, str):
        executable_names = [executable_names]

    local_chrome = get_chrome_download_path()
    _logger.debug(f"Local download path: {local_chrome}")
    if (
        local_chrome.exists()
        and not skip_local
        and local_chrome.stem in executable_names
    ):
        _logger.debug("Returning local chrome")
        return str(local_chrome)
    else:
        _logger.debug(f"Exists? {local_chrome.exists()}")
        _logger.debug(f"Skip local? {skip_local}")
        _logger.debug(
            f"local name: {local_chrome.name} in exe names {executable_names}: "
            f"{local_chrome.name in executable_names}",
        )

    if platform.system() == "Windows":
        os.environ["NoDefaultCurrentDirectoryInExePath"] = "0"  # noqa: SIM112 var name set by windows

    for exe in executable_names:
        if platform.system() == "Windows" and exe == "chrome":
            path = _which_from_windows_reg()
        if path and _is_exe(path):
            return path
        path = shutil.which(exe)
        if path and _is_exe(path):
            return path

    return None


def get_browser_path(*args: Any, **kwargs: Any) -> str | None:  # noqa: D417, don't pass args explicitly
    """
    Call `browser_which()` but check for user override first.

    It looks for the browser in path.

    Accepts the same arguments as `browser_which':

    Args:
        executable_names: the list of names to look for
        skip_local: (default False) don't look for a choreo download of anything.

    """
    return os.environ.get("BROWSER_PATH", browser_which(*args, **kwargs))



from ._async import Broker
from ._sync import BrokerSync

__all__ = [
    "Broker",
    "BrokerSync",
]

# note: should brokers be responsible for closing browser on bad pipe?
# note: should the broker be the watchdog, in that case?



"""
Channels are classes that choreo and the browser use to communicate.

This is a low-level part of the API.

"""

from ._errors import BlockWarning, ChannelClosedError, JSONError
from .pipe import Pipe

__all__ = [
    "BlockWarning",
    "ChannelClosedError",
    "JSONError",
    "Pipe",
]



"""Provide a lower-level sync interface to the Devtools Protocol."""

from __future__ import annotations

from typing import TYPE_CHECKING

import logistro

from choreographer import protocol

if TYPE_CHECKING:
    from typing import Any, MutableMapping

    from choreographer._brokers import BrokerSync


_logger = logistro.getLogger(__name__)


class SessionSync:
    """A session is a single conversation with a single target."""

    session_id: str
    """The id of the session given by the browser."""
    message_id: int
    """All messages are counted per session and this is the current message id."""

    def __init__(self, session_id: str, broker: BrokerSync) -> None:
        """
        Construct a session from the browser as an object.

        A session is like an open conversation with a target.
        All commands are sent on sessions.

        Args:
            broker:  a reference to the browser's broker
            session_id:  the id given by the browser

        """
        if not isinstance(session_id, str):
            raise TypeError("session_id must be a string")
        # Resources
        self._broker = broker

        # State
        self.session_id = session_id
        _logger.debug(f"New session: {session_id}.")
        self.message_id = 0

    def send_command(
        self,
        command: str,
        params: MutableMapping[str, Any] | None = None,
    ) -> Any:
        """
        Send a devtools command on the session.

        https://chromedevtools.github.io/devtools-protocol/

        Args:
            command: devtools command to send
            params: the parameters to send

        Returns:
            A message key (session, message id) tuple or None

        """
        current_id = self.message_id
        self.message_id += 1
        json_command = protocol.BrowserCommand(
            {
                "id": current_id,
                "method": command,
            },
        )

        if self.session_id:
            json_command["sessionId"] = self.session_id
        if params:
            json_command["params"] = params
        _logger.debug(
            f"Sending {command} with {params} on session {self.session_id}",
        )
        return self._broker.write_json(json_command)


class TargetSync:
    """A target like a browser, tab, or others. It sends commands. It has sessions."""

    target_id: str
    """The browser's ID of the target."""
    sessions: MutableMapping[str, SessionSync]
    """A list of all the sessions for this target."""

    def __init__(self, target_id: str, broker: BrokerSync):
        """
        Create a target after one ahs been created by the browser.

        Args:
            broker:  a reference to the browser's broker
            target_id:  the id given by the browser

        """
        if not isinstance(target_id, str):
            raise TypeError("target_id must be string")
        # Resources
        self._broker = broker

        # States
        self.sessions = {}
        self.target_id = target_id
        _logger.debug(f"Created new target {target_id}.")

    def _add_session(self, session: SessionSync) -> None:
        if not isinstance(session, SessionSync):
            raise TypeError("session must be a session type class")
        self.sessions[session.session_id] = session

    def _remove_session(self, session_id: str) -> None:
        if isinstance(session_id, SessionSync):
            session_id = session_id.session_id
        _ = self.sessions.pop(session_id, None)

    def get_session(self) -> SessionSync:
        """Retrieve the first session of the target, if it exists."""
        if not self.sessions.values():
            raise RuntimeError(
                "Cannot use this method without at least one valid session",
            )
        session = next(iter(self.sessions.values()))
        return session

    def send_command(
        self,
        command: str,
        params: MutableMapping[str, Any] | None = None,
    ) -> Any:
        """
        Send a command to the first session in a target.

        https://chromedevtools.github.io/devtools-protocol/

        Args:
            command: devtools command to send
            params: the parameters to send

        """
        if not self.sessions.values():
            raise RuntimeError("Cannot send_command without at least one valid session")
        session = self.get_session()
        return session.send_command(command, params)



import os
import platform

chromium_names = ["chromium", "chromium-browser"]

chrome_names = [
    "chrome",
    "Chrome",
    "google-chrome",
    "google-chrome-stable",
    "Chrome.app",
    "Google Chrome",
    "Google Chrome.app",
    "Google Chrome for Testing",
]
chrome_names.extend(chromium_names)

typical_chrome_paths = None
if platform.system() == "Windows":
    typical_chrome_paths = [
        r"c:\Program Files\Google\Chrome\Application\chrome.exe",
        f"c:\\Users\\{os.environ.get('USER', 'default')}\\AppData\\"
        "Local\\Google\\Chrome\\Application\\chrome.exe",
    ]
elif platform.system() == "Linux":
    typical_chrome_paths = [
        "/usr/bin/google-chrome-stable",
        "/usr/bin/google-chrome",
        "/usr/bin/chrome",
    ]
else:  # assume mac, or system == "Darwin"
    typical_chrome_paths = [
        "/Applications/Google Chrome.app/Contents/MacOS/Google Chrome",
    ]



"""Provides the async api: `Browser`, `Tab`."""

from __future__ import annotations

import asyncio
import os
import subprocess
import warnings
from asyncio import Lock
from typing import TYPE_CHECKING

import logistro

from choreographer import protocol

from ._brokers import Broker
from .browsers import BrowserClosedError, BrowserDepsError, BrowserFailedError, Chromium
from .channels import ChannelClosedError, Pipe
from .protocol.devtools_async import Session, Target
from .utils import TmpDirWarning
from .utils._kill import kill

if TYPE_CHECKING:
    from pathlib import Path
    from types import TracebackType
    from typing import Any, Generator, MutableMapping

    from typing_extensions import Self  # 3.9 needs this, could be from typing in 3.10

    from .browsers._interface_type import BrowserImplInterface
    from .channels._interface_type import ChannelInterface

_logger = logistro.getLogger(__name__)


class Tab(Target):
    """A wrapper for `Target`, so user can use `Tab`, not `Target`."""

    async def close(self) -> None:
        """Close the tab."""
        await self._broker._browser.close_tab(target_id=self.target_id)  # noqa: SLF001


class Browser(Target):
    """`Browser` is the async implementation of `Browser`."""

    tabs: MutableMapping[str, Tab]
    """A mapping by target_id of all the targets which are open tabs."""
    targets: MutableMapping[str, Target]
    """A mapping by target_id of ALL the targets."""
    # Don't init instance attributes with mutables
    _watch_dog_task: asyncio.Task[Any] | None = None

    def _make_lock(self) -> None:
        self._open_lock = Lock()

    async def _is_open(self) -> bool:
        # Did we acquire the lock? If so, return true, we locked open.
        # If we are open, we did not lock open.
        # fuck, go through this again
        if self._open_lock.locked():
            return True
        await self._open_lock.acquire()
        return False

    def _release_lock(self) -> bool:
        try:
            if self._open_lock.locked():
                self._open_lock.release()
                return True
            else:
                return False
        except RuntimeError:
            return False

    def __init__(
        self,
        path: str | Path | None = None,
        *,
        browser_cls: type[BrowserImplInterface] = Chromium,
        channel_cls: type[ChannelInterface] = Pipe,
        **kwargs: Any,
    ) -> None:
        """
        Construct a new browser instance.

        Args:
            path: The path to the browser executable.
            browser_cls: The type of browser (default: `Chromium`).
            channel_cls: The type of channel to browser (default: `Pipe`).
            kwargs: The arguments that the browser_cls takes. For example,
                headless=True/False, enable_gpu=True/False, etc.

        """
        _logger.debug("Attempting to open new browser.")
        self._make_lock()
        self.tabs = {}
        self.targets = {}

        # Compose Resources
        self._channel = channel_cls()
        self._broker = Broker(self, self._channel)
        self._browser_impl = browser_cls(self._channel, path, **kwargs)

    def is_isolated(self) -> bool:
        """Return if process is isolated."""
        return self._browser_impl.is_isolated()

    async def open(self) -> None:
        """Open the browser."""
        _logger.info("Opening browser.")
        if await self._is_open():
            raise RuntimeError("Can't re-open the browser")

        # asyncio's equiv doesn't work in all situations
        if hasattr(self._browser_impl, "logger_parser"):
            parser = self._browser_impl.logger_parser
        else:
            parser = None
        self._logger_pipe, _ = logistro.getPipeLogger(
            "browser_proc",
            parser=parser,
        )

        def run() -> subprocess.Popen[bytes]:
            self._browser_impl.pre_open()
            cli = self._browser_impl.get_cli()
            stderr = self._logger_pipe
            env = self._browser_impl.get_env()
            args = self._browser_impl.get_popen_args()
            return subprocess.Popen(  # noqa: S603
                cli,
                stderr=stderr,
                env=env,
                **args,
            )

        _logger.debug("Trying to open browser.")
        loop = asyncio.get_running_loop()
        self.subprocess = await loop.run_in_executor(None, run)

        super().__init__("0", self._broker)
        self._add_session(Session("", self._broker))

        try:
            _logger.debug("Starting watchdog")
            self._watch_dog_task = asyncio.create_task(self._watchdog())
            _logger.debug("Running read loop")
            self._broker.run_read_loop()
            _logger.debug("Populating Targets")
            await self.populate_targets()
        except (BrowserClosedError, BrowserFailedError, asyncio.CancelledError) as e:
            if (
                hasattr(self._browser_impl, "missing_libs")
                and self._browser_impl.missing_libs
            ):
                raise BrowserDepsError from e
            raise BrowserFailedError(
                "The browser seemed to close immediately after starting.",
                "You can set the `logging.Logger` level lower to see more output.",
                "You may try installing a known working copy of Chrome by running ",
                "`$ choreo_get_chrome`. It may be your copy auto-updated.",
            ) from e

    async def __aenter__(self) -> Self:
        """Open browser as context to launch on entry and close on exit."""
        await self.open()
        return self

    # for use with `await Browser()`
    def __await__(self) -> Generator[Any, Any, Browser]:
        """If you await the `Browser()`, it will implicitly call `open()`."""
        return self.__aenter__().__await__()

    async def _is_closed(self, wait: int | None = 0) -> bool:
        if wait == 0:
            # poll returns None if its open
            _is_open = self.subprocess.poll() is None
            return not _is_open
        else:
            try:
                loop = asyncio.get_running_loop()
                await loop.run_in_executor(None, self.subprocess.wait, wait)
            except subprocess.TimeoutExpired:
                return False
        return True

    async def _close(self) -> None:
        if await self._is_closed():
            _logger.debug("No _close(), already is closed")
            return

        try:
            _logger.debug("Trying Browser.close")
            await self.send_command("Browser.close")
        except (BrowserClosedError, BrowserFailedError):
            _logger.debug("Browser is closed trying to send Browser.close")
            return
        except ChannelClosedError:
            _logger.debug("Can't send Browser.close on close channel")
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, self._channel.close)

        if await self._is_closed(wait=3):
            return

        if await self._is_closed():
            _logger.debug("Browser is closed after closing channel")
            return
        _logger.warning("Resorting to unclean kill browser.")
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, kill, self.subprocess)
        if await self._is_closed(wait=4):
            return
        else:
            raise RuntimeError("Couldn't close or kill browser subprocess")

    async def close(self) -> None:
        """Close the browser."""
        _logger.info("Closing browser.")
        if self._watch_dog_task:
            _logger.debug("Cancelling watchdog.")
            self._watch_dog_task.cancel()
        if not self._release_lock():
            return
        try:
            _logger.debug("Starting browser close methods.")
            await self._close()
            _logger.debug("Browser close methods finished.")
        except ProcessLookupError:
            pass
        self._broker.clean()
        _logger.debug("Broker cleaned up.")
        if self._logger_pipe:
            os.close(self._logger_pipe)
            _logger.debug("Logging pipe closed.")
        self._channel.close()
        _logger.debug("Browser channel closed.")
        self._browser_impl.clean()  # threading this just seems to cause problems
        _logger.debug("Browser implementation cleaned up.")

    async def __aexit__(
        self,
        type_: type[BaseException] | None,
        value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:  # None instead of False is fine, eases type checking
        """Close the browser."""
        await self.close()

    async def _watchdog(self) -> None:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=TmpDirWarning)
            _logger.debug("In watchdog")
            loop = asyncio.get_running_loop()
            _logger.debug2("Running wait.")
            await loop.run_in_executor(None, self.subprocess.wait)
            _logger.warning("Wait expired, Browser is being closed by watchdog.")
            self._watch_dog_task = None
            await self.close()
            await asyncio.sleep(1)
            await loop.run_in_executor(None, self._browser_impl.clean)

    def _add_tab(self, tab: Tab) -> None:
        if not isinstance(tab, Tab):
            raise TypeError(f"tab must be an object of {self._tab_type}")
        self.tabs[tab.target_id] = tab

    def _remove_tab(self, target_id: str) -> None:
        if isinstance(target_id, Tab):
            target_id = target_id.target_id
        del self.tabs[target_id]

    def get_tab(self) -> Tab | None:
        """
        Get the first tab if there is one. Useful for default tabs.

        Returns:
            A tab object.

        """
        if self.tabs.values():
            return next(iter(self.tabs.values()))
        return None

    async def populate_targets(self) -> None:
        """Solicit the actual browser for all targets to add to the browser object."""
        if await self._is_closed():
            raise BrowserClosedError("populate_targets() called on a closed browser")
        response = await self.send_command("Target.getTargets")
        if "error" in response:
            raise RuntimeError("Could not get targets") from Exception(
                response["error"],
            )

        for json_response in response["result"]["targetInfos"]:
            if (
                json_response["type"] == "page"
                and json_response["targetId"] not in self.tabs
            ):
                target_id = json_response["targetId"]
                new_tab = Tab(target_id, self._broker)
                try:
                    await new_tab.create_session()
                except protocol.DevtoolsProtocolError as e:
                    if e.code == protocol.Ecode.TARGET_NOT_FOUND:
                        _logger.warning(
                            f"Target {target_id} not found (could be closed before)",
                        )
                        continue
                    else:
                        raise
                self._add_tab(new_tab)
                _logger.debug(f"The target {target_id} was added")

    async def create_session(self) -> Session:
        """
        Create a browser session. Only in supported browsers, is experimental.

        Returns:
            A session object.

        """
        if await self._is_closed():
            raise BrowserClosedError("create_session() called on a closed browser")
        warnings.warn(  # noqa: B028
            "Creating new sessions on Browser() only works with some "
            "versions of Chrome, it is experimental.",
            protocol.ExperimentalFeatureWarning,
        )
        response = await self.send_command("Target.attachToBrowserTarget")
        if "error" in response:
            raise RuntimeError(
                "Could not create session",
            ) from protocol.DevtoolsProtocolError(
                response,
            )
        session_id = response["result"]["sessionId"]
        new_session = Session(session_id, self._broker)
        self._add_session(new_session)
        return new_session

    async def create_tab(
        self,
        url: str = "",
        width: int | None = None,
        height: int | None = None,
        *,
        window: bool = False,
    ) -> Tab:
        """
        Create a new tab.

        Args:
            url: the url to navigate to, default ""
            width: the width of the tab (headless only)
            height: the height of the tab (headless only)
            window: default False, if true, create new window, not tab

        Returns:
            a tab.

        """
        if await self._is_closed():
            raise BrowserClosedError("create_tab() called on a closed browser.")
        params: MutableMapping[str, Any] = {"url": url}
        if width:
            params["width"] = width
        if height:
            params["height"] = height
        if window:
            params["newWindow"] = True

        response = await self.send_command("Target.createTarget", params=params)
        if "error" in response:
            raise RuntimeError(
                "Could not create tab",
            ) from protocol.DevtoolsProtocolError(
                response,
            )
        target_id = response["result"]["targetId"]
        new_tab = Tab(target_id, self._broker)
        self._add_tab(new_tab)
        await new_tab.create_session()
        return new_tab

    async def close_tab(self, target_id: str) -> protocol.BrowserResponse:
        """
        Close a tab by its id.

        Args:
            target_id: the targetId of the tab to close.

        """
        if await self._is_closed():
            raise BrowserClosedError("close_tab() called on a closed browser")
        if isinstance(target_id, Target):
            target_id = target_id.target_id
        # NOTE: we don't need to manually remove sessions because
        # sessions are intrinsically handled by events
        response = await self.send_command(
            command="Target.closeTarget",
            params={"targetId": target_id},
        )
        self._remove_tab(target_id)
        if "error" in response:
            raise RuntimeError(
                "Could not close tab",
            ) from protocol.DevtoolsProtocolError(
                response,
            )
        return response



from __future__ import annotations

from typing import TYPE_CHECKING

import logistro
import simplejson
from simplejson.errors import JSONDecodeError

from ._errors import JSONError

if TYPE_CHECKING:
    from typing import Any

_logger = logistro.getLogger(__name__)


class MultiEncoder(simplejson.JSONEncoder):
    """Special json encoder for numpy types."""

    def default(self, obj: Any) -> Any:
        if hasattr(obj, "dtype") and obj.dtype.kind == "i" and obj.shape == ():
            return int(obj)
        elif hasattr(obj, "dtype") and obj.dtype.kind == "f" and obj.shape == ():
            return float(obj)
        elif hasattr(obj, "dtype") and obj.shape != ():
            return obj.tolist()
        elif hasattr(obj, "isoformat"):
            return obj.isoformat()
        return simplejson.JSONEncoder.default(self, obj)


def serialize(obj: Any) -> bytes:
    try:
        message = simplejson.dumps(
            obj,
            ensure_ascii=False,
            ignore_nan=True,
            cls=MultiEncoder,
        )
    except JSONDecodeError as e:
        raise JSONError from e
    _logger.debug(f"Serialized: {message[:15]}...{message[-15:]}, size: {len(message)}")
    _logger.debug2(f"Whole message: {message}")

    return message.encode("utf-8")


def deserialize(message: str) -> Any:
    try:
        return simplejson.loads(message)
    except simplejson.errors.JSONDecodeError as e:
        raise JSONError from e



"""Contains implementations of browsers that choreographer can open."""

from ._errors import BrowserClosedError, BrowserDepsError, BrowserFailedError
from .chromium import ChromeNotFoundError, Chromium

__all__ = [
    "BrowserClosedError",
    "BrowserDepsError",
    "BrowserFailedError",
    "ChromeNotFoundError",
    "Chromium",
]



import asyncio

import logistro
import pytest

import choreographer as choreo
from choreographer.protocol import devtools_async

# allows to create a browser pool for tests
pytestmark = pytest.mark.asyncio(loop_scope="function")

_logger = logistro.getLogger(__name__)


# this ignores extra stuff in received- only that we at least have what is expected
def check_response_dictionary(response_received, response_expected):
    for k, v in response_expected.items():
        if isinstance(v, dict):
            check_response_dictionary(v, response_expected[k])
        assert (
            response_received.get(
                k,
                float("NaN"),
            )
            == v
        ), f"Expected: {response_expected}\nReceived: {response_received}"


@pytest.mark.asyncio
async def test_create_and_close_session(browser):
    _logger.info("testing...")
    tab = await browser.create_tab("")
    session = await tab.create_session()
    assert isinstance(session, devtools_async.Session)
    await tab.close_session(session)
    assert session.session_id not in tab.sessions


@pytest.mark.asyncio
async def test_tab_send_command(browser):
    _logger.info("testing...")
    tab = await browser.create_tab("")
    # Test valid request with correct command
    response = await tab.send_command(command="Page.enable")
    check_response_dictionary(response, {"result": {}})

    # Test invalid method name should return error
    response = await tab.send_command(command="dkadklqwmd")
    assert "error" in response

    # Test int method should return error
    with pytest.raises(
        choreo.protocol.MessageTypeError,
    ):
        await tab.send_command(command=12345)


@pytest.mark.asyncio
async def test_subscribe_once(browser):
    _logger.info("testing...")
    tab = await browser.create_tab("")
    subscription_result = tab.subscribe_once("Page.*")
    _ = await tab.send_command("Page.enable")
    _ = await tab.send_command("Page.reload")
    _ = await subscription_result
    assert not subscription_result.exception()


@pytest.mark.asyncio
async def test_subscribe_and_unsubscribe(browser):
    _logger.info("testing...")
    tab = await browser.create_tab("")
    counter = 0
    old_counter = counter

    async def count_event(_r):
        nonlocal counter
        counter += 1

    tab.subscribe("Page.*", count_event)
    assert "Page.*" in next(iter(tab.sessions.values())).subscriptions
    await tab.send_command("Page.enable")
    await tab.send_command("Page.reload")
    await asyncio.sleep(0.5)
    assert counter > old_counter

    tab.unsubscribe("Page.*")
    old_counter = counter

    assert "Page.*" not in next(iter(tab.sessions.values())).subscriptions
    await tab.send_command("Page.enable")
    await tab.send_command("Page.reload")
    assert old_counter == counter



"""
choreographer is a browser controller for python.

choreographer is natively async, so while there are two main entrypoints:
classes `Browser` and `BrowserSync`, the sync version is very limited, functioning
as a building block for more featureful implementations.

See the main README for a quickstart.
"""

from .browser_async import (
    Browser,
    Tab,
)
from .browser_sync import (
    BrowserSync,
    TabSync,
)

__all__ = [
    "Browser",
    "BrowserSync",
    "Tab",
    "TabSync",
]
